// ********************************************************************
//
// constraint_store_headers.h
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// *******************************************************************

#ifndef CONSTRAINT_STORE_HEADERS
#define CONSTRAINT_STORE_HEADERS

#include "constraints_headers.h"
#include "symmetry_headers.h"

// *******************************************************************

// A SolutionReceiver is an object which is prepared to have the individual
// polymorphisms generated by the solving routine sent to itself for
// inspection.

class SolutionReceiver
{
 public:
  virtual ~SolutionReceiver(){}      // Destructor, just in case of subclassing
  typedef SolutionReceiver* CallbackObject;
  typedef void (SolutionReceiver::*CallbackFunction)(const Function&);
};



// *******************************************************************



// The ConstraintStore class is a set of pointers to constraints, along with
// mechanisms for solving the constraints

class ConstraintStore
{
 public:
  // Constructor.  Supply arity and domain of functions
  ConstraintStore(nat a, const std::set<int>& dom);
  
  // Destructor.  Deletes constraints.
  ~ConstraintStore();
  
  // Arity of functions we're generating
  nat get_arity()
    {return arity;}
  
  // Add constraint to store.
  void add_constraint(Constraint* c);

  // Remove constraint from store (true if not found)
  bool remove_constraint(Constraint* c, bool del = true);

  // Output all the constraints on stream s
  void display(std::ostream& s) const;
  
  // Find the polymorphims which satisfy the constraints.  Supply options
  // (their functions are identical to their namesakes' functions in class
  // UserOptions and struct SolverOptions).  Returns number of solution
  // specifications found.
  nat find_polymorphisms(bool quiet,
			 bool stop_at_first,
			 bool force_complete_functions,
			 bool force_all_symmetry_constraints,
			 bool force_essential_constraints,
			 SolutionReceiver::CallbackObject co = NULL,
			 SolutionReceiver::CallbackFunction cf = NULL);

  // Automatically adds certain constraints to the store
  AssociativeConstraint* add_associative_constraint();
  
  CommutativeConstraint* add_commutative_constraint();
  
  IdempotentConstraint* add_idempotent_constraint();
  
  ZeroOneRangeConstraint* add_zero_one_range_constraint();
  
  NearUnanimityConstraint* add_near_unanimity_constraint();

  SquashingConstraint* add_squashing_constraint();

  EssentiallyNAryConstraint* add_essentiallynary_constraint();

  void add_affine_constraint();

  ConservativeConstraint* add_conservative_constraint();

  MaltsevConstraint* add_maltsev_constraint();

  void add_two_semilattice_constraint();

  AutomorphismConstraint* add_automorphism_constraint
               (const RelationStore& relstore,
		bool argument_perms,
		bool da_left,
		bool da_right,
		bool da_both,
		bool accuracy = true);
  
  // For PolymorphismConstraints, fas is a set containing the
  // FunctionArguments restricted by the constraint
  PolymorphismConstraint* add_polymorphism_constraint
    (const std::set<const FunctionArguments*>& fas);

  
  // Exceptions


  // The usual "internal error" exception
  class SanityError{};

  // Thrown if user asks to stop when 1st solution found
  class SolutionExists
    {
    public:
      // Constructor
      SolutionExists(const Function& f)
	{the_solution = f.clone();}

      // Copy constructor
      SolutionExists(const SolutionExists& copy_from)
	{the_solution = (copy_from.the_solution)->clone();}

      // Destructor
      ~SolutionExists()
	{delete(the_solution);}

      // The solution that was found
      Function* the_solution;
    };

  // Thrown if we were asked to find arity-3 commtative fns, etc.
  class ConsistencyError{};

  // Output all constraints on stream s
  friend std::ostream& operator<< (std::ostream &s, const ConstraintStore &cs)
    {cs.display(s); return s;}

 private:

  // The set of constraints in the store.  A superset of...
  std::set <Constraint*> constraints;

  // ...those constraints which need to know when we change search depth
  std::set <Constraint*> notifiable_constraints;

  // Type of iterator over ConstraintStore members
  typedef std::set<Constraint*>::const_iterator CSI;

  // The FunctionArguments we're constraining
  std::set<const FunctionArguments*> constrained_fas;

  // The same, in lexicographic order
  std::vector<const FunctionArguments*> lex_ordering;

  // Arity of the FunctionArguments we're constraining...
  nat arity;

  // ...and their domain
  std::set<int> domain;

  // ********************************************************************
  // All functions and variables below this point are exclusively used by
  // the solving routine.  Most need to be global to the class for speed.
  // ********************************************************************

  struct SolverOptions
  {
    // Don't display solutions
    bool quiet;

    // Stop when we find first solution (throw exception)
    bool stop_at_first;

    // All output functions must be completely specified
    bool force_complete_functions;

    // All output functions must satisfy symmetry constraints
    bool force_all_symmetry_constraints;

    // All output functions must satisfy non-symmetry constraints
    bool force_essential_constraints;
  };

  // Typedef for Constraint List Iterators
  typedef std::list<Constraint*>::iterator CLI;
  
  // The object whose corresponding...
  SolutionReceiver::CallbackObject callback_object;

  // ...function we'll call on finding solutions (or NULL)
  SolutionReceiver::CallbackFunction callback_function;
  
  // Free up resources obtained during the search
  void tidy_up();
  
  // Look for problems in the store. Prints warning or throws SanityError:
  // depends on seriousness
  void check_store_consistency();
  
  // The number of solutions we've found
  nat solution_count;
  
  // Perform the depth-first search for polymorphisms.  Arguments as for
  // apply_constraints(), below.
  void do_search(Function& f,
		 const SolverOptions& options,
		 std::set<const FunctionArguments*>& modified_fas,
		 bool complete);
  
  // Restore all functions, variables, etc., ready for backtracking
  void do_backtrack(Function& f);
  
  // The possible return values of apply_constraints.  Backtrack indicates
  // that we've found a violated constraint, SearchOn means that more
  // searching is required, FoundSolution indicates that we have found a
  // polymorphim specification which satisfies all the constraints.
  enum Action {Backtrack, SearchOn, FoundSolution};
  
  // Apply all not yet-satisfied constraints to the function, where the
  // function values associated with
  // modified_fas have changed since we were last called.  If applying
  // constraints causes other FunctionArguments to change values, then
  // iterate the procedure.  Returns Backtrack if a backtrack needs to be
  // done.
  // On entry, complete is true if, and only if, the function described by
  // f is completely defined (ie., all its FunctionValues sets have size 1).
  // Returns FoundSolution if we have a solution, SearchOn if more
  // searching is required, or Backtrack if no solution is possible on this
  // search branch.  (NB: If the constraints reduce f to being
  // complete, then SearchOn may still be returned, since some constraints
  // may require the 'complete' flag to be set before they can determine
  // whether or not they are satisfied.  Consequently, if the solver spots
  // that f is complete, it should call apply_constraints with the 'complete'
  // flag set to true.)
  Action apply_constraints(Function& f,
			   const SolverOptions& options,
			   std::set<const FunctionArguments*>& modified_fas,
			   bool complete);

  // Current search depth
  nat depth;

  // Maximum search depth
  nat max_depth;
  
  // A lookup table to speed up calculating the FunctionValues corresponding
  // to a singleton domain element.  It only works in favourable (albeit very
  // common) conditions
  std::vector<const FunctionValues*> lookup_singleton_function_values;
  
  // Is the lookup available to us?
  bool lookup_sfv_available;

  
  // The following solver variables are lists rather than sets for speed
  // reasons


  // Constraints which aren't yet satisfied
  std::list<Constraint*> unsatisfied_constraints;
  
  // Constraints which become satisfied at each specific depth
  std::vector< std::list<Constraint*> > satisfied_constraints;
  
   // FAs which need further assignments made to them
  std::list<const FunctionArguments*> unfinished_fas;
  
  // FAs which become redundant at each specific depth
  std::vector< std::list<const FunctionArguments*> > redundant_fas;
};

#endif

