// ***********************************************************************
//
// constraints.cc
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// ***********************************************************************

#include "constraints_headers.h"
#include <cmath>

using namespace std;

// *********************************************************************
// ***        The Constraint class, and its subclasses               ***
// *********************************************************************

// *** NOTE: The implementaion of symmetry-breaking constraints will ***
// *** be found in symmetry.cc                                       ***





// The first version of restrict_range just calls the second version.  This
// is slightly slow, but (probably) the only time it will ever be called is
// at the top level of the search tree, so no harm is done.  If we one day
// write a FunctionValues::integer_to_function_values() routine, we could
// use that here instead.

inline bool Constraint::restrict_range
                        (const FunctionArguments* fa,
			 int d, Function& f,
			 set<const FunctionArguments*>& new_modified_fas) const
{
  set<int> s;
  s.insert(d);
  return restrict_range(fa,
			FunctionValues::set_to_function_values(s),
			f,
			new_modified_fas);
}




// ***************




inline bool Constraint::restrict_range
                        (const FunctionArguments* fa,
			 const FunctionValues* fv,
			 Function& f,
			 set<const FunctionArguments*>& new_modified_fas) const
{
  const FunctionValues* orig_fv = f[fa];
  
  if (fv == orig_fv)
    return false;                                          // Nothing to do

  const FunctionValues* intersect_fv = fv->intersect(orig_fv);
  if (intersect_fv->is_empty_set())
    return true;                                           // Must backtrack

  f.set_value(fa, intersect_fv);
  new_modified_fas.insert(fa);
  return false;
}




// ***************




inline bool Constraint::restrict_to_equal_ranges
                        (const FunctionArguments* fa1,
			 const FunctionArguments* fa2,
			 Function& f,
			 set<const FunctionArguments*>& new_modified_fas) const
{
  const FunctionValues* fv1 = f[fa1];
  const FunctionValues* fv2 = f[fa2];
  
  if (fv1 == fv2)
    return false;                                          // Nothing to do

  const FunctionValues* intersect_fv = fv1->intersect(fv2);
  if (intersect_fv->is_empty_set())
    return true;                                           // Must backtrack

  if (intersect_fv != fv1)
  {
    f.set_value(fa1, intersect_fv);
    new_modified_fas.insert(fa1);
  }
  
  if (intersect_fv != fv2)
  {
    f.set_value(fa2, intersect_fv);
    new_modified_fas.insert(fa2);
  }
  
  return false;
}







// *****************************************************************
// ***              AssociativeConstraint class                  ***
// *****************************************************************


AssociativeConstraint::AssociativeConstraint(nat a, const set<int>& dom)
{
  arity = a;
  domain = dom;
}




// **************************




void AssociativeConstraint::check_consistency
                            (nat a,
			     const set<int>& dom,
			     const set<const FunctionArguments*>& fas) const
{
  if (a != 2)
    throw SanityError();
}




// **************************




Constraint::SatLevel AssociativeConstraint::apply_constraint
                          (Function &f,
			   const set<const FunctionArguments*> &modified_fas,
			   set<const FunctionArguments*> &new_modified_fas,
			   bool complete) const
{
  // We wish to implement the identity f(x,f(y,z)) = f(f(x,y),z).
  //
  // The algoithm used is not particularly clever.  It could be
  // improved by replacing it with the following algorithm.  This (I
  // think) would run slower on small domains, but faster on larger
  // ones.  But I've not implemented it to check.
  //
  //  for all x, y, z, \in domain
  //    Calculate the whole set, S, of values which could be taken by
  //                                    f(f(x,y),z) -- lots of unions
  //    For each possible value, v, of f(y,z)
  //      If the intersection of f(x,v) and S is empty, then
  //        remove v from f(y,z)'s possible vales, and
  //        mark f(y,z) as newly modified
  //    If f(y,z) can take no values now, then return NeverSatisfied
  //    Repeat the last few lines (from "Calculate...") but swapping the
  //                                     roles of LHS and RHS.
  //
  // With a little more thought, it would also be possible to mark
  // f(x,f(y,z)) (and its RHS analogue) as newly-modified under
  // appropriate circumstances (it'd be the union of the clculated
  // intersections, I think).  With even more thought, we could
  // actually make use of the supplied information about which FAs
  // have been recently modified.  I did once do all this thought,
  // but it left me with an algorithm which would have been absolutely
  // horrendous to implement -- as well as very slow to execute.  I
  // don't miss it.

  // The present algorithm only treats the special (read: easy) case
  // when f(y,z) and f(x,y) can only take one value apiece.

  
  // We begin with a few "work variables" which mean different things as we
  // move through the code
  vector<int> tuple(2);
  nat size;
  const FunctionValues* f_values;

  for (set<int>::const_iterator xi = domain.begin();
       xi != domain.end();
       ++xi)
  {
    int x = * xi;
    for (set<int>::const_iterator yi = domain.begin();
	 yi != domain.end();
	 ++yi)
    {
      int y = *yi;
      tuple[0] = x; tuple[1] = y;
      const FunctionArguments* xy =
	FunctionArguments::tuple_to_function_arguments(tuple);
      
      f_values = f[xy];
      size = f_values->get_size();

      // f(x,y) must only take one value
      if (size > 1)
	continue;

      // If it takes no values, we should already have backtracked before now.
      if (size == 0)
	throw SanityError();

      // Find the sole value taken by f(x,y)
      int fxy = *((f_values->function_values_to_set()).begin());

      for (set<int>::const_iterator zi = domain.begin();
	   zi != domain.end();
	   ++zi)
      {
	int z = *zi;
	tuple[0] = y; tuple[1] = z;
	const FunctionArguments* yz =
	  FunctionArguments::tuple_to_function_arguments(tuple);
	
	f_values = f[yz];
	size = f_values->get_size();

	// f(t,z) must only take one value
	if (size > 1)
	  continue;

	// if it takes no values, we should already have backtracked before now
	if (size == 0)
	  throw SanityError();

	// Find the sole value taken by f(y,z)
	int fyz = *((f_values->function_values_to_set()).begin());

	// Now we merely have to ensure that f(x, fyz) and f(fxy, z)
	// share a common range
	tuple[0] = x;
	tuple[1] = fyz;	
	const FunctionArguments* xfyz =
	  FunctionArguments::tuple_to_function_arguments(tuple);
	
	tuple[0] = fxy;
	tuple[1] = z;
	const FunctionArguments* fxyz =
	  FunctionArguments::tuple_to_function_arguments(tuple);
	
	if (restrict_to_equal_ranges(fxyz, xfyz,
				     f, new_modified_fas))
	  return NeverSatisfied;
      }
    }
  }
  
  // Could we manage SometimesSatisfied here?  (Probably not, but...)
  return Unknown;
}




// **************************




void AssociativeConstraint::display(std::ostream& s) const
{
  s << "For all x, y, z, f(x,f(y,z)) = f(f(x,y),z)\n";
}







// *****************************************************************
// ***               CommutativeConstraint class                 ***
// *****************************************************************


void CommutativeConstraint::check_consistency
                            (nat a, const set<int>& dom,
			     const set<const FunctionArguments*>& fas) const
{
  if (a != 2)
    throw SanityError();

  // Check that whenever <x,y> is in fas, then <y,x> is also.
  for (FASI fasi = fas.begin(); fasi != fas.end(); ++fasi)
  {
    if (fas.find( (*fasi)->get_reverse_permutation()) == fas.end())
      throw SanityError();
  }
}




// ******************




Constraint::SatLevel CommutativeConstraint::apply_constraint
                        (Function &f,
			 const set<const FunctionArguments*> &modified_fas,
			 set<const FunctionArguments*> &new_modified_fas,
			 bool complete) const
{
  for (FASI fasi = modified_fas.begin();
       fasi != modified_fas.end();
       ++fasi)
  {
    if (restrict_to_equal_ranges(*fasi, (*fasi)->get_reverse_permutation(),
				 f, new_modified_fas))
      return NeverSatisfied;
  }
  
  return SometimesSatisfied;
}




// *******************




void CommutativeConstraint::display(ostream& s) const
{
  s << "for all x, y, f(x,y) = f(y,x)\n";
}







// *****************************************************************
// ***                  Affine1Constraint class                  ***
// *****************************************************************

// Implements f(f(x,y,z),z,u) = f(x,y,u)
//
// This has similar implementation issues to AssociativeConstraint.  It is
// implemented in a similar way -- that is, by only considering the case
// where f(x,y,z) has size 1.

Affine1Constraint::Affine1Constraint(nat a, const set<int>& dom)
{
  arity = a;
  domain = dom;
}




// *****************




void Affine1Constraint::check_consistency
                        (nat a, const set<int>& dom,
			 const set<const FunctionArguments*>& fas) const
{
  if (a != 3)
    throw SanityError();
}




// *********************




Constraint::SatLevel Affine1Constraint::apply_constraint
                          (Function &f,
			   const set<const FunctionArguments*> &modified_fas,
			   set<const FunctionArguments*> &new_modified_fas,
			   bool complete) const
{
  vector<int> tuple_xyz(3);
  vector<int> tuple_xyu(3);
  vector<int> lhs_tuple(3);

  for (set<int>::const_iterator xi = domain.begin();
       xi != domain.end();
       ++xi)
  {
    int x = *xi;
    tuple_xyz[0] = x;
    tuple_xyu[0] = x;
    
    for (set<int>::const_iterator yi = domain.begin();
	 yi != domain.end();
	 ++yi)
    {
      int y = *yi;
      tuple_xyz[1] = y;
      tuple_xyu[1] = y;
      
      for (set<int>::const_iterator zi = domain.begin();
	   zi != domain.end();
	   ++zi)
      {
	int z = *zi;
	tuple_xyz[2] = z;
	const FunctionArguments* xyz =
	  FunctionArguments::tuple_to_function_arguments(tuple_xyz);
	const FunctionValues* fxyz_fv = f[xyz];
	nat size = fxyz_fv->get_size();

	// We can't do anything unless f(x,y,z) takes exactly one value
	if (size > 1)
	  continue;
	if (size == 0)
	  throw SanityError();

	// Find the value taken by f(x,y,z)
	lhs_tuple[0] = *((fxyz_fv->function_values_to_set()).begin());
	lhs_tuple[1] = z;
	for (set<int>::const_iterator ui = domain.begin();
	     ui != domain.end();
	     ++ui)
	{
	  int u = *ui;
	  lhs_tuple[2] = u;
	  tuple_xyu[2] = u;
	  
	  const FunctionArguments* lhs =
	    FunctionArguments::tuple_to_function_arguments(lhs_tuple);
	  const FunctionArguments* rhs =
	    FunctionArguments::tuple_to_function_arguments(tuple_xyu);
	  
	  if (restrict_to_equal_ranges(lhs, rhs,
				       f, new_modified_fas))
	    return NeverSatisfied;
	}
      }
    }
  }

  // Again, could this be SometimesSatisfied?
  return Unknown;
}




// *********************




void Affine1Constraint::display(ostream& s) const
{
  s << "for all x, y, z, u, f(f(x,y,z),z,u) = f(x,y,u)\n";
}







// *****************************************************************
// ***                  Affine2Constraint class                  ***
// *****************************************************************

// Implements f(x,y,z) = f(z,y,x)

// The alert reader will notice a certain glaring overlap with the
// CommutativeConstraint class.  Actually, even the most half-asleep junior
// doctor at the end of a 72 hour shift in a busy A&E ward which has run
// out of coffee, and who has just taken half-a-dozen valium tables in the
// mistaken belief that they were Pro-Plus, would notice this.  For the sake
// of elegance, the two classes ought to be merged.  But since
// there are only two, I'm not going to bother.

void Affine2Constraint::check_consistency
                        (nat a, const set<int>& dom,
			 const set<const FunctionArguments*>& fas) const
{
  if (a != 3)
    throw SanityError();

  // Check that whenever <x,y,z> is in fas, then <z,y,x> is also.
  for (FASI fasi = fas.begin(); fasi != fas.end(); ++fasi)
  {
    if (fas.find( (*fasi)->get_reverse_permutation()) == fas.end())
      throw SanityError();
  }
}




// ******************




Constraint::SatLevel Affine2Constraint::apply_constraint(
                         Function &f,
			 const set<const FunctionArguments*> &modified_fas,
			 set<const FunctionArguments*> &new_modified_fas,
			 bool complete) const
{
  for (FASI fasi = modified_fas.begin(); fasi != modified_fas.end(); ++fasi)
  {
    if (restrict_to_equal_ranges(*fasi, (*fasi)->get_reverse_permutation(),
				 f, new_modified_fas))
      return NeverSatisfied;
  }
  
  return SometimesSatisfied;
}




// **************************




void Affine2Constraint::display(ostream& s) const
{
  s << "for all x, y, z f(x,y,z) = f(z,y,x)\n";
}







// *****************************************************************
// ***                  Affine3Constraint class                  ***
// *****************************************************************



// Implements f(x,x,y) = y
// Since Affine2 and Affine3 together subsume Maltsev, we could replace
// Affine3 with Maltsev, essentially and say that f(x,x,y) = f(y,x,x) = y.
// But this would not give us any speed increase, as all assignments
// of this form will be done at the top level of the search tree in any case.

void Affine3Constraint::check_consistency
                        (nat a, const set<int>& dom,
			 const set<const FunctionArguments*>& fas) const
{
  if (a != 3)
    throw SanityError();
}




// *********************




Constraint::SatLevel Affine3Constraint::apply_constraint
                         (Function &f,
			  const set<const FunctionArguments*> &modified_fas,
			  set<const FunctionArguments*> &new_modified_fas,
			  bool complete) const
{
  for (FASI fasi = modified_fas.begin(); fasi != modified_fas.end(); ++fasi)
  {
    const FunctionArguments* this_fa = *fasi;
    int a, b; nat bpos;
    
    if (!this_fa->is_almost_constant_function_arguments(a, b, bpos) ||
	(bpos != 2 && a != b))
      continue;
    
    if (restrict_range(this_fa, b, f, new_modified_fas))
      return NeverSatisfied;
  }
  
  return AlwaysSatisfied;
}




// *********************




void Affine3Constraint::display(ostream& s) const
{
  s << "for all x, y, f(x,x,y) = y\n";
}







// ****************************************************************
// ***              IdempotentConstraint class                  ***
// ****************************************************************


void IdempotentConstraint::check_consistency
                           (nat a, const set<int>& dom,
			    const set<const FunctionArguments*>& fas) const
{
  if (a != arity)
    throw SanityError();
}




// ******************




Constraint::SatLevel IdempotentConstraint::apply_constraint
                       (Function &f,
			const set<const FunctionArguments*> &modified_fas,
			set<const FunctionArguments*> &new_modified_fas,
			bool complete) const
{
  if (arity == 0)
    return AlwaysSatisfied;

  for (FASI fasi = modified_fas.begin(); fasi != modified_fas.end(); ++fasi)
  {
    const FunctionArguments* this_fa = *fasi;

    // Nothing to do unless we restrict this FunctionArgument
    if (!(this_fa->is_constant_function_arguments()))
      continue;
    
    if (restrict_range(*fasi, (**fasi)[0], f, new_modified_fas))
      return NeverSatisfied;
  }
  
  return AlwaysSatisfied;
}




// **************




void IdempotentConstraint::display(ostream& s) const
{
  if (arity == 0) return;
  
  s << "for all x, f(";
  for (nat i = 0; i < arity; i++)
  {
    s << "x";
    if (i != arity - 1)
      s << ",";
  }
  s << ") = x\n";
}







// ****************************************************************
// ***              ZeroOneRangeConstraint class                ***
// ****************************************************************


void ZeroOneRangeConstraint::check_consistency
                             (nat a, const set<int>& dom,
			      const set<const FunctionArguments*>& fas) const
{
  if (dom.find(0) == dom.end() || dom.find(1) == dom.end())
    throw SanityError();
}




// ******************




Constraint::SatLevel ZeroOneRangeConstraint::apply_constraint
                           (Function &f,
			    const set<const FunctionArguments*> &modified_fas,
			    set<const FunctionArguments*> &new_modified_fas,
			    bool complete) const
{
  set<int> zero_one;
  zero_one.insert(0);
  zero_one.insert(1);
  const FunctionValues* zero_one_fv =
    FunctionValues::set_to_function_values(zero_one);
  
  for (FASI fasi = modified_fas.begin(); fasi != modified_fas.end(); ++fasi)
  {
    // All constant FAs must take the value 0
    if ((*fasi)->is_constant_function_arguments())
      if (restrict_range(*fasi, 0, f, new_modified_fas))
	return NeverSatisfied;

    // All other FAs can take the value 0 or 1
    if (restrict_range(*fasi, zero_one_fv, f, new_modified_fas))
      return NeverSatisfied;
  }
  return AlwaysSatisfied;
}




// **************




void ZeroOneRangeConstraint::display(ostream& s) const
{
  s << "range(f) is a subset of {0,1}.\n";
  s << "for all x, f(x, x, ..., x) = 0.\n";
}







// **********************************************************************
// ***                NearUnanimityConstraint Class                   ***
// **********************************************************************


void NearUnanimityConstraint::check_consistency
                              (nat a, const set<int>& dom,
			       const set<const FunctionArguments*>& fas) const
{
  if (a != arity)
    throw SanityError();
  
  if (a < 3)
    throw SanityError();
}




// ******************




void NearUnanimityConstraint::display(ostream& s) const
{
  for (nat pos = 0; pos < arity; ++pos)
  {
    s << "for all x, y f(";
    
    for (nat i = 0; i < pos; ++i)
      s << "x";
    s << "y";

    if (pos != arity - 1)
      s << ",";
    
    for (nat i = pos + 1; i < arity; ++i)
    {
      s << "x";
      if (i != arity - 1)
	s << ",";
    }
    
    s << ") = x\n";
  }
}




// *******************




Constraint::SatLevel NearUnanimityConstraint::apply_constraint
                        (Function &f,
			 const set<const FunctionArguments*> &modified_fas,
			 set<const FunctionArguments*> &new_modified_fas,
			 bool complete) const
{
  for (FASI fasi = modified_fas.begin(); fasi != modified_fas.end(); ++fasi)
  {
    const FunctionArguments* this_fa = *fasi;
    int a, b; nat bpos;

    // We only restrict almost-constant FAs
    if (!((this_fa)->is_almost_constant_function_arguments(a, b, bpos)))
      continue;
    
    if (restrict_range(this_fa, a, f, new_modified_fas))
      return NeverSatisfied;
  }

  // We only need to be called at the top level, so we can return
  // AlwaysSatisied
  return AlwaysSatisfied;
}







// *****************************************************************
// ***                ConservativeConstraint class               ***
// *****************************************************************


// Implements f(x_1, x_2, ..., x_n) = x_1 or x_2 or ... or x_n

ConservativeConstraint::ConservativeConstraint(nat a)
{
  arity = a;
}




// *********************




void ConservativeConstraint::check_consistency
                             (nat a, const set<int>& dom,
			      const set<const FunctionArguments*>& fas) const
{
  if (arity != a)
    throw SanityError();
}




// *********************




Constraint::SatLevel ConservativeConstraint::apply_constraint
                        (Function &f,
			 const set<const FunctionArguments*> &modified_fas,
			 set<const FunctionArguments*> &new_modified_fas,
			 bool complete) const
{
  for (FASI fasi = modified_fas.begin(); fasi != modified_fas.end(); ++fasi)
  {
    set<int> s;
    const FunctionArguments* this_fa = *fasi;
    vector<int> this_fa_tuple = this_fa->function_arguments_to_tuple();
    
    copy(this_fa_tuple.begin(),this_fa_tuple.end(), inserter(s, s.end()));
    const FunctionValues* allowed_fv =
      FunctionValues::set_to_function_values(s);
    
    if (restrict_range(this_fa, allowed_fv, f, new_modified_fas))
      return NeverSatisfied;
  }
  
  return AlwaysSatisfied;
}




// *********************




void ConservativeConstraint::display(ostream& s) const
{
  switch (arity)
  {
  case 0:
    break;
  case 1:
    s << "For all x, f(x) = x\n";
    break;
  case 2:
    s << "For all x, y, f(x,y) \\in {x,y}\n";
    break;
  case 3:
    s << "For all x, y, z, f(x,y,z) \\in {x,y,z}\n";
    break;
  default:
    s << "For all ";

    for (nat i = 1; i <= arity; ++i)
      s << "x_" << i << ", ";
    s << "f(";

    for (nat i = 1; i <= arity; ++i)
    {
      s << "x_" << i;
      if (i < arity)
	s << ",";
    }
    s << ") \\in {";

    for (nat i = 1; i <= arity; ++i)
    {
      s << "x_" << i;
      if (i < arity)
	s << ",";
    }
    s << "}\n";
  }
}







// *****************************************************************
// ***               TwoSemilattice1Constraint class             ***
// *****************************************************************


// Implements f(x,f(y,x)) = f(y,x)

// As usual for these "nested application of f" constraints, we choose
// the simplest method of implementation, and only deal with the case
// where f(y,x) can take exactly one value.  A better algorithm almost
// certainly exists.  See AssociativeConstraint for inspiration.

TwoSemilattice1Constraint::TwoSemilattice1Constraint(nat a,
						     const set<int>& dom)
{
  arity = a;
  domain = dom;
}




// *********************




void TwoSemilattice1Constraint::check_consistency
                               (nat a, const set<int>& dom,
				const set<const FunctionArguments*>& fas) const
{
  if (a != 2)
    throw SanityError();
  
}

// *********************

Constraint::SatLevel TwoSemilattice1Constraint::apply_constraint
                        (Function &f,
			 const set<const FunctionArguments*> &modified_fas,
			 set<const FunctionArguments*> &new_modified_fas,
			 bool complete) const
{
  vector<int> tuple_yx(2);
  vector<int> tuple_lhs(2);
  
  for (set<int>::const_iterator xi = domain.begin();
       xi != domain.end();
       ++xi)
  {
    int x = *xi;
    tuple_yx[1] = x;
    tuple_lhs[0] = x;
    for (set<int>::const_iterator yi = domain.begin();
	 yi != domain.end();
	 ++yi)
    {
      int y = *yi;
      tuple_yx[0] = y;
      const FunctionArguments* yx =
	FunctionArguments::tuple_to_function_arguments(tuple_yx);
      const FunctionValues* fyx_fv = f[yx];
      nat size = fyx_fv->get_size();

      // f(y,x) can only take one value, otherwise we dont know how to restrict
      // other function values.
      if (size > 1)
	continue;
      if (size == 0)
	throw SanityError();

      // Get the sole value taken by f(y,x)
      int fyx = *((fyx_fv->function_values_to_set()).begin());
      tuple_lhs[1] = fyx;
      const FunctionArguments* lhs =
	FunctionArguments::tuple_to_function_arguments(tuple_lhs);
      
      // May as well pass fyx_fv as second argument, since restrict_range
      // would otherwise only have to convert fyx back to a FunctionValues*.
      // So we save a bit of time.
      if (restrict_range(lhs, fyx_fv, f, new_modified_fas))
	return NeverSatisfied;
    }
  }
  return Unknown;            // SometimesSatisfied?
}




// *********************




void TwoSemilattice1Constraint::display(ostream& s) const
{
  s << "for all x, y, f(x,f(y,x)) = f(y,x)\n";
}







// *****************************************************************
// ***                  MaltsevConstraint class                  ***
// *****************************************************************


// Implements f(x,y,y) = f(y,y,x) = x

void MaltsevConstraint::check_consistency
                        (nat a, const set<int>& dom,
			 const set<const FunctionArguments*>& fas) const
{
  if (a != 3)
    throw SanityError();
}




// *********************




Constraint::SatLevel MaltsevConstraint::apply_constraint
                         (Function &f,
			  const set<const FunctionArguments*> &modified_fas,
			  set<const FunctionArguments*> &new_modified_fas,
			  bool complete) const
{
  for (FASI fasi = modified_fas.begin(); fasi != modified_fas.end(); ++fasi)
  {
    const FunctionArguments* this_fa = *fasi;
    int a, b; nat bpos;

    // Do we constrain this FA?
    // (NB: Genuinely constant tuples return bpos == 0 here, so the following
    // line deals with that case fine)
    if (!(this_fa->is_almost_constant_function_arguments(a, b, bpos)) ||
	bpos == 1)
      continue;

    // If so, restrict the range of its values appropriately
    if (restrict_range(this_fa, b, f, new_modified_fas))
      return NeverSatisfied;
  }

  // We only need to be called at the top search level, so we can return
  // AlwaysSatisfied
  return AlwaysSatisfied;
}




// *********************




void MaltsevConstraint::display(ostream& s) const
{
  s << "for all x, y, f(x,y,y) = f(y,y,x) = x\n";
}







// **********************************************************************
// ***                SquashingConstraint Class                       ***
// **********************************************************************


// Forces arity-1 functions to not be injective.
// Since there are so few arity-1 functions, we can afford to be lazy in
// implementing this constraint.  We recalculate 'range' at each search
// depth, rather than storing it somewhere, and making ourselves notifiable.

SquashingConstraint::SquashingConstraint(const set<int>& dom)
{
  domain = dom;
}




// *********************




void SquashingConstraint::check_consistency
                          (nat a, const set<int>& dom,
			   const set<const FunctionArguments*>& fas) const
{
  if (a != 1)
    throw SanityError();
}




// *********************




Constraint::SatLevel SquashingConstraint::apply_constraint
                         (Function &f,
			  const set<const FunctionArguments*> &modified_fas,
			  set<const FunctionArguments*> &new_modified_fas,
			  bool complete) const
{
  set<int> range;
  range.clear();

  // Determine the range of the part of the function which is completely
  // defined
  for (set<int>::const_iterator si = domain.begin(); si != domain.end(); ++si)
  {
    vector<int> vec (1);
    vec[0] = *si;
    const FunctionArguments* dom_e =
      FunctionArguments::tuple_to_function_arguments(vec);
    const FunctionValues* range_fv = f[dom_e];
    
    if (range_fv->get_size() != 1)
      continue;
    
    // If f(x) and f(y) both map to the same value (and x != y), then the
    // function is not injective -- ie., it is a squashing function.
    int range_e = *(range_fv->function_values_to_set().begin());
    if (!(range.insert(range_e).second))
      return AlwaysSatisfied;
  }

  if (complete)
    return NeverSatisfied;
  
  // All incomplete function specifications can define at least one
  // squashing function
  return SometimesSatisfied;
}




// **********************




void SquashingConstraint::display(ostream& s) const
{
  s << "f is not injective\n";
}








// *************************************************************************
// ***                   PolymorphismConstraint Class                    ***
// *************************************************************************

PolymorphismConstraint::PolymorphismConstraint
                        (const set <const FunctionArguments*>& fas,
			 const set<int>& dom,
			 nat arity)
{
  restricted_fas = fas;
  vector<int> empty_vector;
  
  for (FASI fasi = fas.begin(); fasi != fas.end(); ++fasi)
    possible_value_vectors[*fasi] = empty_vector;
  
  possible_value_vectors_size = 0;
  available_indices.clear();
  
  // Copy the set into a vector, for passing to Function constructor
  vector<const FunctionArguments*> fav;
  copy(fas.begin(), fas.end(), back_inserter(fav));
  possible_value_function = new(Function)(arity, dom, fav, true);
  
  // Assign the empty set to each FunctionArgument we constrain.  (The
  // above initialisation sets each FA to be the whole of dom.)
  set<int> emptyset;
  const FunctionValues* emptyset_fv =
    FunctionValues::set_to_function_values(emptyset);
  
  for (FAVI favi = fav.begin(); favi != fav.end(); ++favi)
    possible_value_function->set_value(*favi, emptyset_fv);
}




// ******************




PolymorphismConstraint::~PolymorphismConstraint()
{
  delete(possible_value_function);
}




// ******************




void PolymorphismConstraint::add_possible_assignments
                       (set<pair<const FunctionArguments*, int> > possibility)
{
  // First check that the disjunction supplied is not self-contradictory.
  // The way to do this is to ensure that <fa,i> pair which occurs in
  // possibility has a unique value of fa.
  set<const FunctionArguments*> fas_seen;
  for (set<pair<const FunctionArguments*, int> >::const_iterator
	 pi = possibility.begin();
       pi != possibility.end();
       ++pi)
  {
    if ((fas_seen.insert(pi->first)).second == false)
      return;
  }
  
  // We now know that all the FAs in possibility are distinct, and the
  // possibility is safe to add as a disjunct
  for (set<pair<const FunctionArguments*, int> >::const_iterator
	 pi = possibility.begin();
       pi != possibility.end();
       ++pi)
  {
    map<const FunctionArguments*, vector<int> >::iterator insertion_point =
      possible_value_vectors.find(pi->first);

    // We're stuck if an FA we don't know about is offered as a possibility
    if (insertion_point == possible_value_vectors.end())
      throw SanityError();
    
    (insertion_point->second).push_back(pi->second);

    // Add pi->second to the set of values which we already know might be
    // taken by pi->first
    set<int> values =
     (possible_value_function->get_value(pi->first))->function_values_to_set();
    
    values.insert(pi->second);
    possible_value_function->set_value(pi->first, FunctionValues::set_to_function_values(values));
  }

  // The size of each vector has now increased.  Update the record of the
  // size, and mark the new index as a valid possibility
  available_indices.push_back(possible_value_vectors_size);
  possible_value_vectors_size++;
}




// ******************




Constraint::SatLevel PolymorphismConstraint::apply_constraint
                        (Function &f,
			 const set<const FunctionArguments*> &modified_fas,
			 set<const FunctionArguments*> &new_modified_fas,
			 bool complete) const
{
  // The following should never happen unless our constraint was empty to
  // begin with.
  if (available_indices.size() == 0)
    return NeverSatisfied;

  // Determine if we've removed any available indices
  bool done_erasure = false;
  for (FASI fasi = modified_fas.begin(); fasi != modified_fas.end(); ++fasi)
  {
    const FunctionArguments* this_fa = *fasi;
 
    map<const FunctionArguments*, vector<int> >::const_iterator this_fa_location =
      possible_value_vectors.find(this_fa);

    // If we don't constrain this FA, then continue
    if (this_fa_location == possible_value_vectors.end())
      continue;

    const FunctionValues* f_values = f[this_fa];
    const FunctionValues* our_values =
      possible_value_function->get_value(this_fa);

    // If we can't eliminate any vector indices, then continue
    if (f_values == our_values)
      continue;

    const FunctionValues* intersection_fv = f_values->intersect(our_values);
    if (intersection_fv->is_empty_set())
      return NeverSatisfied;
    
    // Two possibilities: either f[this_fa]'s possible values can allow us
    // to erase some indices from available_indices, or else our knowledge
    // of available_indices will allow us to restrict f[this_fa]'s possible
    // values (or both).

    // First, try restricting the available indices to being those which
    // contain values in intersection_fv

    // Get the vector of values we currently allow.
    const vector<int>& this_fa_values = this_fa_location->second;

    // This is really a for-loop, but we're modifying iterators as we move
    // through it
    list<nat>::iterator indexi = available_indices.begin();
    while (indexi != available_indices.end())
    {
      nat index = *indexi;
      // Can we eliminate the index?
      if (intersection_fv->is_element(this_fa_values[index]))
      {
	// No.
	++indexi;
	continue;
      }

      // Yes.  Eliminate the index from the list of available indices
      done_erasure = true;
      list<nat>::iterator next_indexi = ++indexi;
      unavailable_indices[depth].splice(unavailable_indices[depth].begin(),
					available_indices,
					--indexi);
      indexi = next_indexi;
    }
    
    // (Erasure done, if necessary)
    new_modified_fas.insert(this_fa);
    f.set_value(this_fa, intersection_fv);
  }

  // If we've erased any available indexes, update f and possible_value
  // function appropriately
  if (done_erasure)
  {
    for (map<const FunctionArguments*, vector<int> >::const_iterator
	   mapi = possible_value_vectors.begin();
	 mapi != possible_value_vectors.end();
	 ++mapi)           // For each FA we restrict
    {
      const FunctionArguments* fa = mapi->first;
      set<int> values;
      for (list<nat>::const_iterator indexi = available_indices.begin();
	   indexi != available_indices.end();
	   ++indexi)
      {
	values.insert(mapi->second[*indexi]);
      }
      
      const FunctionValues* new_fv =
	FunctionValues::set_to_function_values(values);

      if (new_fv->is_empty_set())                  // This should never happen
	return NeverSatisfied;

      if (possible_value_function->get_value(fa) != new_fv)
      {
	possible_value_function->set_value(fa, new_fv);
	f.set_value(fa, new_fv);
	new_modified_fas.insert(fa);
      }
    }
  }


  // Now we must determine if the constraint is always satisfied.  It is
  // always satisfied if either:
  //   (a) There is only one available index left, or
  //   (b) The remaining possibilities for the restricted FunctionArguments
  //       form an "allows-anything" relation over some domain.
  // (Of course, (a) is just a special case of (b).)
  if (available_indices.size() == 1)
    return AlwaysSatisfied;

  const FunctionValues* first_values =
    possible_value_function->get_value((possible_value_vectors.begin())->first);
  for (map<const FunctionArguments*, vector<int> >::const_iterator
	 mapi = possible_value_vectors.begin();
       mapi != possible_value_vectors.end();
       ++mapi)
  {
    if (possible_value_function->get_value(mapi->first) != first_values)
      return SometimesSatisfied;
  }
  
  // All the FAs can take the same set of values.
  if ((double) available_indices.size() == 
      (double) pow((float) (first_values->get_size()),
		   (int) possible_value_vectors.size()))
  {
    return AlwaysSatisfied;
  }
  
  return SometimesSatisfied;
}




// ******************




void PolymorphismConstraint::notify_search_starting(nat max_d)
{
  depth = 0;
  max_depth = max_d;

  // Resize unavailable_indices, and make sure it's empty
  unavailable_indices.resize(max_depth+1);
  for (nat i = 0; i < max_depth + 1; ++i)
    if (unavailable_indices[i].size() != 0)
      throw SanityError();
}



// ********************



void PolymorphismConstraint::notify_search_descent()
{
  possible_value_function->save_state();

  depth++;
}




// *********************




void PolymorphismConstraint::notify_search_ascent()
{
  possible_value_function->restore_state();
  
  // Put back any indices which we had said were unavailable
  available_indices.splice(available_indices.begin(),
			   unavailable_indices[depth]);

  depth--;
}




// *****************




void PolymorphismConstraint::notify_search_ending()
{
  // Do nothing
}    




// ******************




void PolymorphismConstraint::display(ostream& s) const
{
  // The choice between the two varieties of for-loop is the choice between
  // whether or not we want constraints to be updated dynamically as we
  // eliminate possible indices.  This makes no difference to the end-user,
  // but when debugging, we sometimes want one behaviour, and sometimes the
  // other.

  //  for (nat i = 0; i < possible_value_vectors_size; ++i)
  for (list<nat>::const_iterator li = available_indices.begin();
       li != available_indices.end();
       ++li)
  {
    nat i = *li;   // DELETE if using the other for-loop.
    if (i != 0)
      s << "|| ";
    
    s << "(";
    
    for (map<const FunctionArguments*, vector<int> >::const_iterator
	   mapi = possible_value_vectors.begin();
	 mapi != possible_value_vectors.end();
	 ++mapi)
    {
      if (mapi != possible_value_vectors.begin())
	s << " && ";
      
      s << "f(";
      display_tuple(s, (mapi->first)->function_arguments_to_tuple());
      s << ") = ";
      s << (mapi->second)[i];
    }
    s << ")\n";
  }
  s << "-------\n";
}




// ******************
