// ********************************************************************
//
// constraints_headers.h
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// *******************************************************************

#ifndef CONSTRAINTS_HEADERS

#define CONSTRAINTS_HEADERS

#include <set>
#include <list>
#include <algorithm>
#include "function_headers.h"

// ********************************************************************

// The Constraint class just details the public interface to constraints
// in general.  All constraints are objects of subclasses of the Constraint
// class.

class Constraint
{
 public:
  virtual ~Constraint()
    {}
  // The various types of constraints we have implemented.
  enum ConstraintType
    {Associative, Commutative, Idempotent, NearUnanimity, Polymorphism,
     Affine1, Affine2, Affine3, Conservative, TwoSemilattice1, Maltsev,
     EssentiallyNAry, Squashing, ZeroOneRange, Automorphism};
  
  // How satisfied a constraint on a function specification can be.
  enum SatLevel
    {Unknown, SymmetryBreaking, NeverSatisfied, SometimesSatisfied,
     AlwaysSatisfied};
  
  // Whether a constraint constrains a given FunctionArgument.  (Order is
  // important.)
  enum Constrainedness
    {Unconstrained, SymmetryConstrained, Constrained};
  
  // The type of the present constraint
  virtual ConstraintType get_constraint_type() const = 0;
  
  // Can we have the argument-permutation automorphism constraint invoked
  // at the same time as this constraint?
  virtual bool permutation_closed() const = 0;
  
  // Do we need to be notified when the search depth changes?  (Only needed
  // if we have internal data structs)
  virtual bool need_level_notification() const = 0;
  
  // Does the constraint affect f(*fa)?
  virtual Constrainedness
    is_constraint_relevant(const FunctionArguments* fa) const = 0;
  
  // Check that the constraint agrees with its store re: arity, domain
  // size, and constrained tuples.
  // Print a warning on cerr for minor problems; throw SanityError for
  // major problems.
  virtual void check_consistency
                (nat a,
		 const std::set<int>& dom,
		 const std::set<const FunctionArguments*>& fas) const
    {return;}
  
  // Apply the constraint to the function, assuming that modified_fas
  // have changed values since we last applied this constraint.  Any
  // FunctionArguments which we in turn change (including those from
  // modified_fas) should be placed in new_modified_fas.  Returns the
  // "satisfaction level" of the resulting function specification.
  // Should endeavour to return NeverSatisfied where appropriate, but
  // if we forget, the solver will notice later in any case (as long
  // as complete is false -- and if it's true, a warning message will
  // be printed when the function is output).  If complete is true on
  // entry then the function f is completely defined on every
  // FunctionArguments tuple which we constrain (according to our
  // setting of is_constraint_relevant() ).
  virtual SatLevel apply_constraint
                     (Function &f,
		      const std::set<const FunctionArguments*> &modified_fas,
		      std::set<const FunctionArguments*> &new_modified_fas,
		      bool complete) const = 0;
  
  // Display the constraint in plain English
  virtual void display(std::ostream& s) const = 0;

  // ***********************************************************
  // Some members which are only called if we say we need search
  // depth notification.
  // ***********************************************************

  // About to start search.  Max search depth should be supplied
  virtual void notify_search_starting(nat max_d) {};
  
  // Moving down a search level
  virtual void notify_search_descent() {};

  // Moving up a search level
  virtual void notify_search_ascent() {};

  // Search is about to end
  virtual void notify_search_ending() {};

  // ************
  //  Exceptions
  // ************

  // A SanityError is thrown if internal sanity check is violated
  class SanityError{};

  // *********
  //  Friends
  // *********

  // Outputting a constraint on a stream
  friend std::ostream& operator<< (std::ostream &s, const Constraint &c)
    {c.display(s); return s;}
  
 protected:

  // We begin with a couple of typdefs.  These are purely to save typing
  // FunctionArgumentsVectorIterator and FunctionArgumentsSetIterator.
  typedef std::vector<const FunctionArguments*>::const_iterator FAVI;
  typedef std::set<const FunctionArguments*>::const_iterator FASI;
  
  // Sets f[fa] = f[fa] \intersect {d}.  Returns true iff this leaves empty
  // set.  Adds fa to new_modified_fas if necessary.  This is merely a special
  // case of the other version of restrict_range(), below.
  bool restrict_range
             (const FunctionArguments* fa,
	      int d,
	      Function& f,
	      std::set<const FunctionArguments*>& new_modified_fas) const;
  
  // Sets f[fa] = f[fa] \intersect fv.  Returns true iff this leaves empty
  // set.  Adds fa to new_modified_fas if necessary
  bool restrict_range
             (const FunctionArguments* fa,
	      const FunctionValues* fv,
	      Function& f,
	      std::set<const FunctionArguments*>& new_modified_fas) const;
  
  // Sets f[fa1] = f[fa2] = f[fa1] \intersect f[fa2].  Returns true iff
  // this leaves empty set.  Adds fa1 and/or fa2 to new_modified_fas if
  // necessary
  bool restrict_to_equal_ranges
             (const FunctionArguments* fa1,
	      const FunctionArguments* fa2,
	      Function& f,
	      std::set<const FunctionArguments*>& new_modified_fas) const;
};





// ********************************************************************
//                       THE CONSTRAINTS
// ********************************************************************





// An AssociativeConstraint ensures that an arity-2 function is associative

class AssociativeConstraint : public Constraint
{
 public:
  AssociativeConstraint(nat a, const std::set<int>& dom);

  ConstraintType get_constraint_type() const
    {return Associative;}

  bool permutation_closed() const
    {return true;}

  bool need_level_notification() const
    {return false;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {return Constrained;}

  void check_consistency(nat a,
			 const std::set<int>& dom,
			 const std::set<const FunctionArguments*>& fas) const;

  SatLevel apply_constraint
                  (Function &f,
		   const std::set<const FunctionArguments*> &modified_fas,
		   std::set<const FunctionArguments*> &new_modified_fas,
		   bool complete) const;

  void display(std::ostream& s) const;
  
 private:
  nat arity;
  std::set<int> domain;
};





// ********************************************************************





// A CommutativeConstraint ensures that an arity-2 function is commutative.
// On functions of arity other than 2, it has no effect.  We implement
// commutativity as one large constraint rather than as many smaller ones
// (of the form f<1,2> = f<2,1>; f<3,4> = f<4,3>, etc.)

class CommutativeConstraint : public Constraint
{
 public:
  ConstraintType get_constraint_type() const
    {return Commutative;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {
      if (fa->is_constant_function_arguments())
	return Unconstrained;
      else
	return Constrained;
    }
  
  // Commutative functions are those unchanged under permutation.
  bool permutation_closed() const
    {return true;}

  bool need_level_notification() const
    {return false;}

  void check_consistency(nat a,
			 const std::set<int>& dom,
			 const std::set<const FunctionArguments*>& fas) const;

  SatLevel apply_constraint
                 (Function &f,
		  const std::set<const FunctionArguments*> &modified_fas,
		  std::set<const FunctionArguments*> &new_modified_fas,
		  bool complete) const;

  void display(std::ostream& s) const;
};






// *********************************************************************






// An IdempotentConstraint ensures that an arity-a function is idempotent.
// It is implemented as one large constraint, rather than many small ones.

class IdempotentConstraint : public Constraint
{
 public:
  IdempotentConstraint(nat a)                 // Constructor.  Supply arity.
    {arity = a;}

  ConstraintType get_constraint_type() const
    {return Idempotent;}

  bool permutation_closed() const
    {return true;}

  bool need_level_notification() const
    {return false;}


  void check_consistency(nat a,
			 const std::set<int>& dom,
			 const std::set<const FunctionArguments*>& fas) const;
  
  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {
      if ((arity > 0) && (fa->is_constant_function_arguments()))
	return Constrained;
      else
	return Unconstrained;
    }

  SatLevel apply_constraint
          (Function &f,
	   const std::set<const FunctionArguments*> &modified_fas,
	   std::set<const FunctionArguments*> &new_modified_fas,
	   bool complete) const;

  void display(std::ostream& s) const;
  
 private:
  
  // The arity of functions we apply this constraint to
  nat arity;
};






// *********************************************************************





// A ZeroOneRangeConstraint simply ensures that
//   (a) the range of any output function is a subset of {0,1}; and
//   (b) f(any constant tuple) = 0.
// It is typically used when Polyanna is used to enumerate all relations over
// a domain of a fixed size (>=2).
// It is implemented as one large constraint, rather than many small ones.

class ZeroOneRangeConstraint : public Constraint
{
 public:
  ConstraintType get_constraint_type() const
    {return ZeroOneRange;}

  bool permutation_closed() const
    {return true;}

  bool need_level_notification() const
    {return false;}


  void check_consistency(nat a,
			 const std::set<int>& dom,
			 const std::set<const FunctionArguments*>& fas) const;
  
  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {return Constrained;}

  SatLevel apply_constraint
              (Function &f,
	       const std::set<const FunctionArguments*> &modified_fas,
	       std::set<const FunctionArguments*> &new_modified_fas,
	       bool complete) const;

  void display(std::ostream& s) const;
};






// ***********************************************************************




// The NearUnanimityConstraint forces functions to be closed under a near-
// unanimity operation.

class NearUnanimityConstraint : public Constraint
{
 public:
  // Constructor.  Supply arity.
  NearUnanimityConstraint(nat a)
    {arity = a;}

  ConstraintType get_constraint_type() const
    {return NearUnanimity;}

  bool permutation_closed() const
    {return true;}

  bool need_level_notification() const
    {return false;}

  void check_consistency(nat a,
			 const std::set<int>& dom,
			 const std::set<const FunctionArguments*>& fas) const;
  
  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {
      int a,b;
      nat bp;
      if (fa->is_almost_constant_function_arguments(a, b, bp))
	return Constrained;
      else
	return Unconstrained;
    }

  SatLevel apply_constraint
            (Function &f,
	     const std::set<const FunctionArguments*> &modified_fas,
	     std::set<const FunctionArguments*> &new_modified_fas,
	     bool complete) const;

  void display(std::ostream& s) const;
  
 private:
  // The arity of functions we apply this constraint to
  nat arity;
};





// *********************************************************************




// The Affine1Constraint is one of the three constraints which between them
// ensure that a function is affine.
// In particular, the Affine1Constraint ensures that f(f(x,y,z),z,u) = f(x,y,z)

class Affine1Constraint : public Constraint
{
 public:
  Affine1Constraint(nat a, const std::set<int>& dom);

  ConstraintType get_constraint_type() const
    {return Affine1;}

  bool permutation_closed() const
    {return false;}

  bool need_level_notification() const
    {return false;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {return Constrained;}

  void check_consistency(nat a,
			 const std::set<int>& dom,
			 const std::set<const FunctionArguments*>& fas) const;

  SatLevel apply_constraint
              (Function &f,
	       const std::set<const FunctionArguments*> &modified_fas,
	       std::set<const FunctionArguments*> &new_modified_fas,
	       bool complete) const;

  void display(std::ostream& s) const;

 private:
  nat arity;
  std::set<int> domain;
};






// *********************************************************************





// The Affine2Constraint is one of the three constraints which between them
// ensure that a function is affine.
// In particular, the Affine2Constraint ensures that f(x, y, z) == f(z, y, x)

class Affine2Constraint : public Constraint
{
 public:
  ConstraintType get_constraint_type() const
    {return Affine2;}

  bool permutation_closed() const
    {return false;}

  bool need_level_notification() const
    {return false;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {
      if (fa == fa->get_reverse_permutation())
	return Unconstrained;
      else
	return Constrained;
    }

  void check_consistency
            (nat a,
	     const std::set<int>& dom,
	     const std::set<const FunctionArguments*>& fas) const;

  SatLevel apply_constraint
            (Function &f,
	     const std::set<const FunctionArguments*> &modified_fas,
	     std::set<const FunctionArguments*> &new_modified_fas,
	     bool complete) const;

  void display(std::ostream& s) const;
};






// *********************************************************************





// The Affine3Constraint is one of the three constraints which between them
// ensure that a function is affine.
// In particular, the Affine3Constraint ensures that f(x, x, y) == y

class Affine3Constraint : public Constraint
{
 public:
  ConstraintType get_constraint_type() const
    {return Affine3;}

  bool permutation_closed() const
    {return false;}

  bool need_level_notification() const
    {return false;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {
      int a, b;
      nat bp;
      if (fa->is_almost_constant_function_arguments(a, b, bp) &&
	  (bp == 2 || a == b))
	return Constrained;
      else
	return Unconstrained;
    }

  void check_consistency
          (nat a,
	   const std::set<int>& dom,
	   const std::set<const FunctionArguments*>& fas) const;

  SatLevel apply_constraint
          (Function &f,
	   const std::set<const FunctionArguments*> &modified_fas,
	   std::set<const FunctionArguments*> &new_modified_fas,
	   bool complete) const;

  void display(std::ostream& s) const;
};







// *********************************************************************





// The ConservatyiveConstraint class ensures that any fucntion to which it
// is applied is conservative

class ConservativeConstraint : public Constraint
{
 public:
  ConservativeConstraint(nat a);

  ConstraintType get_constraint_type() const
    {return Conservative;}

  bool permutation_closed() const
    {return true;}

  bool need_level_notification() const
    {return false;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {return Constrained;}

  void check_consistency
         (nat a,
	  const std::set<int>& dom,
	  const std::set<const FunctionArguments*>& fas) const;

  SatLevel apply_constraint
         (Function &f,
	  const std::set<const FunctionArguments*> &modified_fas,
	  std::set<const FunctionArguments*> &new_modified_fas,
	  bool complete) const;

  void display(std::ostream& s) const;

 private:
  // The function arity we're searching for
  nat arity;
};







// *********************************************************************





// The TwoSemilattice1Constraint class ensures that f(x, f(y, x)) == f(y, x).
// Together with the CommutativeConstraint and IdempotentConstraint classes,
// this constraint restricts arity-2 functions to being 2-semilattices.

class TwoSemilattice1Constraint : public Constraint
{
 public:
  // Constructor.  Supply arity and domain
  TwoSemilattice1Constraint(nat a, const std::set<int>& dom);

  ConstraintType get_constraint_type() const
    {return TwoSemilattice1;}

  bool permutation_closed() const
    {return true;}                           // Actually false, but only used in conjunction with commutative

  bool need_level_notification() const
    {return false;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {return Constrained;}

  void check_consistency
         (nat a,
	  const std::set<int>& dom,
	  const std::set<const FunctionArguments*>& fas) const;

  SatLevel apply_constraint
         (Function &f,
	  const std::set<const FunctionArguments*> &modified_fas,
	  std::set<const FunctionArguments*> &new_modified_fas,
	  bool complete) const;

  void display(std::ostream& s) const;

 private:
  nat arity;
  std::set<int> domain;
};






// *********************************************************************





// The MaltsevConstraint class is applied to functions of arity 3.  It
// imp[lements the constraint that f(x, y, y) == f(y, x, x) == x.  That is,
// That f is a Mal'tsev function.

class MaltsevConstraint : public Constraint
{
 public:
  ConstraintType get_constraint_type() const
    {return Maltsev;}

  bool permutation_closed() const
    {return false;}

  bool need_level_notification() const
    {return false;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {
      int a, b;
      nat bp;
      if ((fa->is_almost_constant_function_arguments(a, b, bp)) && bp != 1)
	return Constrained;
      else
	return Unconstrained;
    }

  void check_consistency
        (nat a,
	 const std::set<int>& dom,
	 const std::set<const FunctionArguments*>& fas) const;

  SatLevel apply_constraint
         (Function &f,
	  const std::set<const FunctionArguments*> &modified_fas,
	  std::set<const FunctionArguments*> &new_modified_fas,
	  bool complete) const;

  void display(std::ostream& s) const;
};







// ***********************************************************************







// A SquashingConstraint restricts arity-1 functions to being non-injective

class SquashingConstraint : public Constraint
{
 public:
  // Constructor.  Supply domain.
  SquashingConstraint(const std::set<int>& dom);
  
  ConstraintType get_constraint_type() const
    {return Squashing;}

  bool permutation_closed() const
    {return true;}

  bool need_level_notification() const
    {return false;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {return Constrained;}

  void check_consistency
        (nat a,
	 const std::set<int>& dom,
	 const std::set<const FunctionArguments*>& fas) const;

  SatLevel apply_constraint
        (Function &f,
	 const std::set<const FunctionArguments*> &modified_fas,
	 std::set<const FunctionArguments*> &new_modified_fas,
	 bool complete) const;

  void display(std::ostream& s) const;

 private:
  std::set<int> domain;
};






// ************************************************************************







// Between them, the PolymorphismConstraints ensure that any functions
// output are polymorphims of the input relations.  Each constraint defines
// a disjunction of conjunctions and says, intuitively, that f(this tuple
// from R, that tuple from R, the other tuple from R) = first tuple from R OR
// f(this tuple from R, that tuple from R, the other tuple from R) = second
// tuple from R OR ... f(this tuple from R, that tuple from R, the other
// tuple from R) = last tuple from R
// Thus, each relation of size s gives rise to s^a PolymorphismConstraints
// (each of which is a disjunction of size s), where a is the arity of
// polymorphism we're searching for.

class PolymorphismConstraint : public Constraint
{
 public:
  // Constructor.  Supply FAs we affect (as a vector, for convenience),
  // and their domain, and their arity
  PolymorphismConstraint(const std::set<const FunctionArguments*>& fas,
			 const std::set<int>& dom,
			 nat arity);

  // Destructor
  ~PolymorphismConstraint();

  // Adds one possible disjunct to the constraint
  void add_possible_assignments
	    (std::set<std::pair<const FunctionArguments*, int> > possibility);

  bool permutation_closed() const
    {return true;}

  bool need_level_notification() const
    {return true;}

  ConstraintType get_constraint_type() const
    {return Polymorphism;}

  Constrainedness is_constraint_relevant(const FunctionArguments* fa) const
    {
      if (restricted_fas.find(fa) == restricted_fas.end())
	return Unconstrained;
      else
	return Constrained;
    }

  SatLevel apply_constraint
               (Function &f,
		const std::set<const FunctionArguments*> &modified_fas,
		std::set<const FunctionArguments*> &new_modified_fas,
		bool complete) const;


   // About to start search.  Max search depth must be passed as argument
  void notify_search_starting(nat max_d);

  // Moving down a search level
  void notify_search_descent();
  
  // Moving up a search level
  void notify_search_ascent();

  // Search ending
  void notify_search_ending();    

  void display(std::ostream& s) const;

 private:
  // The FunctionArguments we restrict
  std::set <const FunctionArguments*> restricted_fas;
  
  // Possible values taken by each FunctionArguments.  The vectors are in
  // the same order for each FA -- that is, each index into the vector gives
  // one allowable simultaneous assignment to all the FunctionArguments in
  // the map.
  std::map<const FunctionArguments*, std::vector<int> > possible_value_vectors;
  
  // The length of each vector in possible_values
  nat possible_value_vectors_size;
  
  // Indices into above vectors which have not yet been ruled out as
  // impossible. This is always a subset of [0, ..., possible_values_size)
  mutable std::list<nat> available_indices;
  
  // The indices which become unavailable at each search depth
  mutable std::vector<std::list<nat> > unavailable_indices;
  
  // This does the same job as possible_value_vectors, but only keeps track
  // of the *set* of values which can currently be taken by any
  // FunctionArgument.  It is implemented as a partial function, for speed.
  Function* possible_value_function;
  
  // Current search depth
  nat depth;
  
  // Maximum search depth
  nat max_depth;
};

#endif
