// ************************************************************************
//
// function_headers.h
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// ************************************************************************

#ifndef FUNCTION_HEADERS

#define FUNCTION_HEADERS

#include <iostream>
#include <vector>
#include <set>
#include <map>
#include <utility>
#include "misc_headers.h"

class FunctionArguments;
class FunctionValues;
class Function;

// ************************************************************************
// ***                    FunctionArguments class                       ***
// ************************************************************************

// An object of class FunctionArguments is just a tuple -- the
// arguments to a function.  The tuple is augmented with several
// useful operations, such as, "Is the tuple constant?", "What happens
// if you permute the tuple members in a standard order?", etc.
// FunctionArguments objects are automatically created when
// needed, via tuple_to_function_arguments().  This ensures that
// each tuple is only created once as a FunctionArguments object --
// ensuring efficiency.

class FunctionArguments
{
 public:

  // Output routine
  void display(std::ostream &s) const;

  // All perms of a given FunctionArgument
  void display_perms() const;

  // The arity of the tuple
  nat get_arity() const
    {return tuple.size();}

  // Access to tuple
  int operator[] (nat n) const
    {return tuple[n];}

  // Access to whole tuple
  std::vector<int> function_arguments_to_tuple() const
    {return tuple;}

  // Converts a tuple (vector) to a FunctionArguments object.  It constructs
  // the object if needed
  static const FunctionArguments* tuple_to_function_arguments
                                        (const std::vector<int> &t);

  // Number of permutations of this FA.  Should always equal arity! - 1
  nat get_number_of_permutations() const
    {return permutations.size();}

  // Return i+1'th non-trivial permutation
  const FunctionArguments* get_permutation(nat i) const
    {return permutations[i];}

  // Return ourselves, reversed.
  const FunctionArguments* get_reverse_permutation() const
    {return rev_perm;}

  // Is the tuple <a,a,a,...,a>?
  bool is_constant_function_arguments() const
    {return constant_function_arguments;}

  
  // Is tuple <a,a,...,a,b,a,...,a>? (arity > 2).  If so, values of a & b,
  // and posn of b are returned
  bool is_almost_constant_function_arguments(int& a, int& b, nat& bpos) const;

  // ************
  // Exceptions
  // ************

  // A SanityError is thrown if ever an event occurs whcih should not logically
  // be able to happen.  It indicates a programming bug.
  class SanityError{};

 private:

  // The tuple we represent
  std::vector<int> tuple;

  // The map to from tuples to fn_arguments
  static std::map<std::vector<int>, const FunctionArguments*> tuple_to_fn_args_map;

  // Not a true constructor, for technical reasons
  void FunctionArguments_constructor(const std::vector<int> &t);

  // Permutations in fixed order (inc. duplicates)
  std::vector<const FunctionArguments*> permutations;

   // Create permutations list
  void generate_permutations();

  // Index of reverse permutation (-1 if none)
  const FunctionArguments* rev_perm;

  // For is_constant_function_arguments()
  bool constant_function_arguments;

  // Set up constant_function_arguments
  void initialise_constant_function_arguments();

  // For is_almost_constant_function_arguments()
  bool almost_constant_function_arguments;

  // The value of the almost-constant part
  int acfa_major_value;

  // The value of the remaining argument...
  int acfa_minor_value;

  // ... and its position in the tuple
  nat acfa_minor_posn;

  // Set up the above variables
  void initialise_almost_constant_function_arguments();

  // ****************************************************************
  // Interfacing with the Function class.  Slightly dangerously, the
  // Function objects have access to the whole representation of
  // FunctionArguments.  It was either that, or make some of the
  // following functions fully public.
  // ****************************************************************

  friend class Function;

  // Automatically increased when necessary
  static nat max_functions_definable;

  
  static std::set<nat> function_ids_available;

  // Construct new function.
  static nat get_new_function_id();

  // Destroy function.
  static void destroy_function(nat id);

  // Change number of functions we can define
  void resize_functions_definable(nat new_size) const
    {apply_fn_result.resize(new_size);}

  // fa.(afr[id]) == f_id[fa]
  mutable std::vector<const FunctionValues*> apply_fn_result;

  // Get f_id[fa]
  const FunctionValues* get_function_result(nat id) const
    {return apply_fn_result[id];}

  // Set f_id[fa]
  void set_function_result(nat id, const FunctionValues* fv) const
    {apply_fn_result[id] = fv;}


  // **********************************************************************
  //       Outputting FunctionArguments
  // **********************************************************************

  // Normla output on a stream
  friend std::ostream& operator<< (std::ostream& s,
				   const FunctionArguments& fa)
    {
      fa.display(s);
      return s;
    }
  
  // Output via pointer.  This is a convenience in many cases.  It saves a
  // dereference in the main program.
  friend std::ostream& operator<< (std::ostream& s,
				   const FunctionArguments* fa)
    {
      fa->display(s);
      return s;
    }
};









// ************************************************************************
// ***                    FunctionValues class                          ***
// ************************************************************************




// A FunctionValues object is a specification of the possible values taken by
// a function when applied to a tuple.  Normally, it is a set of
// integers.  As in the case of FunctionArguments, each set is only
// created once.  This enables constant-time comparisons of set equality:
// Two sets are equal iff they have the same pointer address.

class FunctionValues
{
 public:
  // Just in case we subclass with a destructor
  virtual ~FunctionValues(){}

  // Converts a set of ints to a FunctionValues object.  Constructs the object
  // if needed
  static const FunctionValues* set_to_function_values(const std::set<int> &s);

  // Goes the other way, regaining the corresponding set of values
  std::set<int> function_values_to_set() const
    {return values;}

  // Output the set of values
  void display(std::ostream& s) const
   {display_set(s, values);}

  // Are we the empty set?  (Function provided for efficiency)
  virtual bool is_empty_set() const
    {return false;}

  // The minimum value in the set
  virtual int min_value() const
    {return min_val;}

  // The maximum value in the set
  virtual int max_value() const
    {return max_val;}

   // The size of the set
  virtual nat get_size() const
    {return size_of_set;}

  // Is e a member of the set we represent?
  virtual bool is_element(int e) const
    {return values.find(e) != values.end();}

  
  // Intersection of ourself and some other FunctionValues set
  virtual const FunctionValues* intersect(const FunctionValues* other_fv) const
    {return find_intersection(this, other_fv);}

  // ************
  //  Exceptions
  // ************

  // Thrown if the set of values becomes empty
  class EmptySet{};

  // A sanity check has been violated
  class SanityError{};

 private:

  // The fake "constructor"
  virtual void FunctionValues_constructor(const std::set<int> &s);

  /// The map to fn_values
  static std::map<std::set<int>, const FunctionValues*> set_to_fn_values_map;

  // The set itself
  std::set<int> values;

  
  // Find the intersection of the two FunctionValues sets.  Uses a
  // table lookup for speed.
  static const FunctionValues* find_intersection(const FunctionValues* fv1,
						 const FunctionValues* fv2);

  // The table used by the previous function
  static std::map<std::pair<const FunctionValues*, const FunctionValues*>,
                  const FunctionValues*> intersection_table;

  //The minimum value in the set
  int min_val;

  // and the maximum value
  int max_val;

  // The size of the set
  nat size_of_set;


  // ************************************
  //   Outputting FunctionValues
  // ************************************

  // Normal output of a set correcponding to the FunctionValues supplied
  friend std::ostream& operator<< (std::ostream& s, const FunctionValues& fv)
    {
      fv.display(s);
      return s;
    }

  // Output of the same, except the FunctionValues is supplied as a pointer.
  // This saves a dereferewnce in the main code.
  friend std::ostream& operator<< (std::ostream& s, const FunctionValues* fv)
    {
      fv->display(s);
      return s;
    }
};





// A version of the class used when the set of values is empty.
// The only reason it's split off into a separate class is for efficiency.

class EmptyFunctionValues : public FunctionValues
{
 public:
  bool is_empty_set() const
    {return true;}
  
  int min_value() const
    {throw EmptySet();}
  
  int max_value() const
    {throw EmptySet();}
  
  nat get_size() const
    {return 0;}
  
  bool is_element(int e) const
    {return false;}
  
  const FunctionValues* intersect(const FunctionValues* other_fv) const
    {return this;}

 private:
  void FunctionValues_constructor(const std::set<int> &s);
  
  // values is always {}.  It's included to save rewriting functions like
  // display()
  std::set<int> values;
};






// ************************************************************************
// ***                       Function class                             ***
// ************************************************************************




// Each object is a function specification, defining one or more
// functions.  At its heart, a function specification is a map from
// (pointers to) FunctionArguments to (pointers to) a FunctionValues set.
// An actual map is not used, for speed reasons.  Instead, we interface
// closely with the FunctionArguments class, which is really where functions
// are stored.  The Function class is basically just syntactic sugar to
// disguise this.

class Function
{
 public:
  // Constructor.  Supply arity, domain, and default ordering of
  // FunctionArguments for output
  Function(nat a,
	   const std::set<int>& dom,
	   const std::vector<const FunctionArguments*>& ord,
	   bool partial = false);

  // Destructor
  ~Function();

  // Create a copy of the function
  Function* clone() const;
  
  // Output routine.  Doesn't display the function very prettily
  void display_as_list
            (std::ostream& s,
	     const std::vector<const FunctionArguments*>& ordering) const;

  // Displays using default tuple ordering.
  void display_as_list(std::ostream& s) const
    {display_as_list(s, default_ordering);}

  // Output routine.  Displays fn as a relation.  Tuple t is output iff
  // f(t) != {0}.  This is mainly only useful when using Polyanna to
  // calculate essentially different relations.
  void display_as_relation
            (std::ostream& s,
	     const std::vector<const FunctionArguments*>& ordering) const;

  // Displays using default tuple ordering.
  void display_as_relation(std::ostream& s) const
    {display_as_relation(s, default_ordering);}
  
  // Access to the definition of f.  Returns pointer to FValues of f(*fa)
  const FunctionValues* get_value(const FunctionArguments* fa) const
    {return fa->get_function_result(function_id);}

  // Abbreviation for above
  const FunctionValues* operator[] (const FunctionArguments* fa) const
    {return get_value(fa);}

  // f(fa) = fv.
  void set_value(const FunctionArguments* fa, const FunctionValues* fv)
    {
      history.push_back(std::make_pair(fa, get_value(fa)));
      fa->set_function_result(function_id, fv);
    }

  // **********************
  //  Backtrack management
  // **********************

  // Save current function state
  void save_state()
    {history.push_back(history_separator);}

  // Restore to previously saved state.
  void restore_state();

  
  // ************
  //  Exceptions
  // ************

  // Thrown if internal error detected
  class SanityError{};

  // Thrown if too many calls made to restore_state()
  class Underflow{};

 private:

  // The id obtained from FunctionArguments::gnfid()
  nat function_id;

  // Arity of the function
  nat arity;

  // function:(domain^arity) --> Pow(domain)
  std::set<int> domain;

  // Default ordering of FunctionArguments for fn o/p
  std::vector<const FunctionArguments*> default_ordering;

  // Is the function partial?
  bool is_partial;

  // Backtrack history
  std::vector<std::pair<const FunctionArguments*, const FunctionValues*> >
    history;

  // save_state point
  static std::pair<const FunctionArguments*, const FunctionValues*>
    history_separator;


  // ************************************
  //      Outputting functions
  // ************************************

  // Output on a stream
  friend std::ostream& operator<< (std::ostream& s, const Function& f)
    {f.display_as_list(s); return s;}
};

#endif
