// ***************************************************************************
//
// polyanna.cc
//
// part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// ***************************************************************************

// *******************************
//
//     Polyanna main program
//         Richard Gault
//             2001
//
// *******************************

#include "front_end_headers.h"
#include "relation_headers.h"
#include <iostream>
#include <string>
#include <cstring>

using namespace std;

const nat major_version = 1;
const nat minor_version = 01;



// ***********************************************************
// ***                 UserOptions Class                   ***
// ***********************************************************

UserOptions::UserOptions()
{
  fn_arity = 2;
  perm_opts = true;
  automorphisms = true;
  essentiallynary = true;
  associative = false;
  commutative = false;
  idempotent = false;
  near_unanimity = false;
  affine = false;
  conservative = false;
  twosemilattice = false;
  maltsev = false;
  squash = false;
  force_essential_constraints = true;
  force_all_symmetry_constraints = false;
  force_complete_functions = false;
  force_level = 1;
  stop_at_first = false;
  quiet = false;
  count = true;
  tractable = false;
  view_constraints = false;
  view_relations = false;
  enumerate_relations = -1;
  iterations = 1;
}




// **************




// Who needs getopt.h? :-)
// Written by hand, largely because not everyone has the getopt headers
// available, and this program will be made available on the web.

nat UserOptions::read_options(nat argc, char* argv[])
{
  if (argc == 1)                     // User ran the program with no arguments
    display_short_help();
  
  // Is long help required?  Needed so we don't print help twice
  bool display_help = false;
  
  nat i;
  for (i = 1; i < argc; ++i)
  {
    string option = argv[i];
    
    if (option.length() == 0)
      throw SanityError();
    
    if (option[0] != '-')    // Not an option.  Assume it's a filename.
      break;
    
    option.erase(0,1);

    if (option.length() == 0)
      throw UnknownOption("<empty option>");
    
    bool long_form;
    if (option[0] == '-')
    {
      long_form = true;
      option.erase(0,1);
      if (option.length() == 0)
	throw UnknownOption("<empty option>");
    }
    else
      long_form = false;

    bool next_option_used;
    bool next_opt_from_next_opt;

    // Loop to extract several options out of a short-form list: eg. -kta3
    do
    {
      bool flag = true;
      while (option.length() > 0 && option[0] == '~')
      {
	flag = !flag;
	option.erase(0,1);
      }
      
      if (option.length() == 0)                // The null option was supplied
	throw UnknownOption("<empty option>");
      
      // If long_form is true, set up next_option to be either the number
      // at the end of the option (eg. --arity7 sets next_option to "7"),
      // or else the next option on the command line.  If long_form is false,
      // set up next_option to be the rest of the option ("-a7") or else
      // the next option on the command line.
      string next_option = "";
      next_opt_from_next_opt = false;              // the "-arity 7" case
      next_option_used = false;                    // Must be set in the "switch" statement below
      bool next_option_numeric;                    // Does next_option contain a number? ("" -> false)
      int next_option_value = 0;                   // If so, what is is?
      
      if (long_form)
      {
	for (nat j = 0; j < option.length(); ++j)
	  if ((option[j] >= '0' && option[j] <= '9') || option[j] == '-')
	  {
	    next_option = option.substr(j);
	    option.erase(j);
	    break;
	  }
	
	if (option.length() == 0)                 // Unexpected numeric optio
	  throw UnknownOption(next_option);
      }
      else
      {
	next_option = option.substr(1);
	option.erase(1);
      }
      
      if (next_option == "" && i < argc - 1)
      {
	next_option = argv[i+1];
	next_opt_from_next_opt = true;
      }
      
      next_option_numeric = true;
      for (nat j = 0; j < next_option.length(); ++j)
	if (!((next_option[j] >= '0' && next_option[j] <= '9') ||
	      (j == 0 && next_option[j] == '-')))
	{
	  next_option_numeric = false;
	  break;
	}
      
      if (next_option == "")
	next_option_numeric = false;
      
      if (next_option_numeric)
	next_option_value = atoi(next_option.c_str());
      
      if (next_option == "0" ||
	  (long_form && next_opt_from_next_opt && next_option == "off") ||
	  (long_form && next_opt_from_next_opt && next_option == "no") ||
	  (long_form && next_opt_from_next_opt && next_option == "n"))
      {
	// Don't set flag = false, in case user types "-~idempotent no" which,
	// logically, should set flag to true.  Probably.
	flag = !flag;
	next_option_used = true;
      }
      
      if (next_option == "1" ||
	  (long_form && next_opt_from_next_opt && next_option == "on") ||
	  (long_form && next_opt_from_next_opt && next_option == "yes") ||
	  (long_form && next_opt_from_next_opt && next_option == "y"))
      {
	//flag = flag;
	next_option_used = true;
      }

      // ******************************************************
      // The code to handle the individual options begins here.
      // ******************************************************

      // Any command line which needs the value from next_option should
      // set next_option_used to be true
      if ((long_form && option == "arity") || (!long_form && option == "a"))
      {
	if (!next_option_numeric || next_option_value < 0)
	  throw MalformedOption("arity");
	fn_arity = next_option_value;
	next_option_used = true;
      }
      
      else if ((long_form && option == "permopts") ||
	       (long_form && option == "permutationoptimisations") ||
	       (!long_form && option == "p"))
	perm_opts = flag;
      
      else if ((long_form && option == "essentiallynary") ||
	       (!long_form && option == "e"))
	essentiallynary = flag;
      
      else if ((long_form && option == "automorphisms") ||
	       (!long_form && option == "m"))
	automorphisms = flag;
      
      else if ((long_form && option == "associative") ||
	       (!long_form && option == "A"))
      {
	// If the user changes the arity later, an error will be issued
	associative = flag;
	if (flag) fn_arity = 2;
      }
      
      else if ((long_form && option == "commutative") ||
	       (!long_form && option == "C"))
	{
	  // It's still possible user might turn perms off later:
	  // A warning will be issued if they are.
	  commutative = flag;
	  fn_arity = 2;
	  if (flag) perm_opts = false;
	}
	
      else if ((long_form && option == "idempotent") ||
	       (!long_form && option == "I"))
	idempotent = flag;

      // No direct short form for --semilattice
      else if (long_form && (option == "semilattice" || option == "ACI"))
      {
	associative = flag;
	commutative = flag;
	idempotent = flag;
	if (flag) perm_opts = false;
	if (flag) fn_arity = 2;
      }
      
      else if ((long_form &&
		(option == "majority" ||
		 option == "nuf" ||
		 option == "nearunanimity")) ||
	       (!long_form && option == "M"))
      {
	// We don't enforce that "majority" means arity == 3
	near_unanimity = flag;
	if (flag && fn_arity < 3)
	  fn_arity = 3;
      }
      
      else if ((long_form && option == "force") ||
	       (!long_form && option == "f"))
      {
	// The force value must be between 0 and 3
	if (!next_option_numeric ||
	    next_option_value < 0 ||
	    next_option_value > 3)
	  throw MalformedOption("force");
	
	next_option_used = true;
	force_level = next_option_value;
	
	switch (force_level)
	{
	case 0:
	  force_essential_constraints = false;
	  force_all_symmetry_constraints = false;
	  force_complete_functions = false;
	  break;
	  
	case 1:
	  force_essential_constraints = true;
	  force_all_symmetry_constraints = false;
	  force_complete_functions = false;
	  break;
	  
	case 2:
	  force_essential_constraints = true;
	  force_all_symmetry_constraints = true;
	  force_complete_functions = false;
	  break;
	  
	case 3:
	  force_essential_constraints = true;
	  force_all_symmetry_constraints = true;
	  force_complete_functions = true;
	  break;
	  
	default:
	  throw SanityError();
	}
      }
      
      else if ((long_form && option == "affine") ||
	       (!long_form && option == "F"))
      {
	affine = flag;
	if (flag) fn_arity = 3;
	if (flag) perm_opts = false; // Affine is not closed under permutations
      }
      
      else if ((long_form && option == "conservative") ||
	       (!long_form && option == "T"))
	conservative = flag;
      
      else if ((long_form && (option == "maltsev" || option == "mal'tsev")) ||
	       (!long_form && option == "V"))
      {
	maltsev = flag;
	if (flag) fn_arity = 3;
	if (flag) perm_opts = false;;
      }
      
      else if ((long_form && option == "reduce") ||
	       (!long_form && option == "r") ||
	       (long_form && option == "squash"))
	squash = flag;
      
      else if ((long_form && option == "twosemilattice") ||
	       (!long_form && option == "S"))
      {
	twosemilattice = flag;
	if (flag) fn_arity = 2;
	if (flag) perm_opts = false;
      }
      
      else if ((long_form && option == "stoponfirst") ||
	       (!long_form && option == "s"))
	stop_at_first = flag;
      
      else if ((long_form && option == "quiet") ||
	       (!long_form && option == "q"))
	quiet = flag;
      
      else if ((long_form && option == "count") ||
	       (!long_form && option == "c"))
	count = flag;
      
      else if ((long_form && option == "tractable") ||
	       (!long_form && option == "t"))
	tractable = true;
      
      else if ((long_form && option == "help") ||
	       (!long_form && option == "h"))
	display_help = true;
      
      else if ((long_form && option == "viewconstraints") ||
	       (long_form && option == "vc"))                 // No short form
	view_constraints = flag;
      
      else if ((long_form && option == "viewrelations") ||
	       (long_form && option == "vr"))                 // No short form
	view_relations = flag;
      
      else if (long_form && option == "solve")                // No short form
	iterations = flag ? 1 : 0;
      
      else if (long_form && option == "iterations")
      {
	if (!next_option_numeric)
	  throw MalformedOption("iterations");
	
	iterations = next_option_value;
	next_option_used = true;
	quiet = true;
	count = false;
      }
      
      else if (long_form &&
	       (option == "enumeraterelations" || option == "er"))
      {
	if (!next_option_numeric || next_option_value < 0)
	  throw MalformedOption("enumeraterelations");
	
	enumerate_relations = next_option_value;
	next_option_used = true;
      }

      // If we get this far, then none of the command line arguments have
      // matched.
      else
	throw UnknownOption(option);

      if (next_option_used && next_opt_from_next_opt)
	++i;
      
      if (!long_form && !next_option_used && !next_opt_from_next_opt)
	option = next_option;
      
    } while (!long_form &&
	     !next_option_used &&
	     !next_opt_from_next_opt &&
	     option.length() > 0);         // End of do-loop
  }

  if (display_help)
    display_long_help();
  
  // Some sanity checks before we leave
  if (quiet && !count && !tractable && !view_constraints &&
      !view_relations && i < argc)
    cerr << "Warning: You have selected no output to be displayed\n";

  return i;
}




// ******************




void UserOptions::display_short_help() const
{
  cout << "Polyanna " << major_version << ".";
  if (minor_version < 10)
    cout << "0";
  cout << minor_version << " by Richard Gault, 2001-2004\n";
  cout << "Copyright (C) The University of Oxford, 2001-2004.\n\n";
  cout << "Polyanna is a program for finding the polymorphisms of a set of relations.\n";
  cout << "Use \"polyanna -h\" for a summary of the command-line options.  For full\n";
  cout << "details, see the user guide.\n\n";
}




// ******************




void UserOptions::display_long_help() const
{
  cout << "Polyanna " << major_version << ".";
  if (minor_version < 10)
    cout << "0";
  cout << minor_version << " by Richard Gault, 2001\n";
  cout << "Copyright (C) The University of Oxford, 2001, 2002.\n\n";
  cout << "Polyanna is a program for finding the polymorphisms of a set of relations.\n";
  cout << "The following is only a brief outline of the command-line options available:\n";
  cout << "for full details, see the user guide.\n\n";
  cout << "Usage:\n";
  cout << "       polyanna [options] [<input-file> [<input-file> ...]]\n\n";
  cout << "-a <n> or --arity <n>\n";
  cout << "       Search for polymorphisms of arity <n>.  For default, see manual.\n";
  cout << "-p or --permopts\n";
  cout << "       Display permutations once only.  For default, see manual.\n";
  cout << "-e or --essentiallynary\n";
  cout << "       All functions must make essential use of each of their arguments (ON).\n";
  cout << "-m or --automorphisms\n";
  cout << "       Factor out symmetries due to automorphisms.\n";
  cout << "-A or --associative\n";
  cout << "       All functions must be associative.  (Arity 2 only.)\n";
  cout << "-C or --commutative\n";
  cout << "       All functions must be commutative.  (Arity 2 only.)\n";
  cout << "-I or --idempotent\n";
  cout << "       All functions must be idempotent.\n";
  cout << "--semilattice (or --ACI)\n";
  cout << "       An alternative way of writing -A -C -I (or -ACI).\n";
  cout << "-M or --majority or --nearunanimity (or --nuf)\n";
  cout << "       All functions must be near-unanimity functions.  (Arity 3 or more only.)\n";
  cout << "-F or --affine\n";
  cout << "       All functions must be affine.  (Arity 3 only.)\n";
  cout << "-V or --maltsev (or --mal'tsev)\n";
  cout << "       All functions must be Mal'tsev.  (Arity 3 only.)\n";
  cout << "-T or --conservative\n";
  cout << "       All functions must be conservative.\n";
  cout << "-S or --twosemilattice\n";
  cout << "       All functions must be 2-semilattice functions.  (Arity 2 only.)\n";
  cout << "-r or --reduce (or --squash).\n";
  cout << "       Convert the relations to a reduced set, by applying squashing function.\n";
  cout << "-f <n> or --force <n>\n";
  cout << "       How hard do we search for functions?  See the manual.  Default is 1.\n";
  cout << "-s or --stoponfirst\n";
  cout << "       Stop as soon as we find the first solution.\n";
  cout << "-q or --quiet\n";
  cout << "       Don't display the solutions as we find them.\n";
  cout << "-c or --count\n";
  cout << "       Display the number of function specifications found (ON).\n";
  cout << "-t or --tractable\n";
  cout << "       Determine whether the input relations are tractable.\n";
  cout << "-h or --help\n";
  cout << "       Display this help information.\n";
  cout << "--viewconstraints or --vc\n";
  cout << "       Display all the constraints just before we begin searching.\n";
  cout << "--viewrelations or --vr\n";
  cout << "       Display the decomposition of all the input relations.\n";
  cout << "--solve\n";
  cout << "       Perform the polymorphism search.  Default is ON, of course.\n";
  cout << "--iterations <n>\n";
  cout << "       Perform the same search <n> times (for timing).  Default is 1.\n";
  cout << "--enumeraterelations <n> or --er <n>\n";
  cout << "       Generate all essentially different relations over a domain of size n.\n";
  cout << "\nAll Boolean options are OFF by default, except where indicated otherwise.\n";
  cout << "To explicitly turn off an option which is ON by default use, say, \"-e0\".\n\n";
}

void UserOptions::display_options_summary() const
{
  if (tractable)
  {
    cout << "(Testing for tractability.)\n";
    return;
  }
  
  cout << "(Arity " << fn_arity << " ";
  if (essentiallynary)
    cout << "(essential)";
  else
    cout << "(inessential)";
  
  cout << ", force level " << force_level;
  
  if (perm_opts)
    cout << ", functions listed once up to permutation,\n";
  else
    cout << ", all permutations listed,\n";

  if (automorphisms)
    cout << " automorphisms factored out,\n";
  
  if (associative || commutative || idempotent || affine || near_unanimity ||
      conservative || twosemilattice || maltsev)
  {
    cout << " functions restricted to being ";
    
    if (associative && commutative && idempotent)
      cout << "semilattice ";
    else
    {
      if (associative) cout << "associative ";
      if (commutative) cout << "commutative ";
      if (idempotent)  cout << "idempotent ";
    }
    
    if (affine)
      cout << "affine ";
    
    if (near_unanimity && fn_arity == 3)
     cout << "majority ";
    
    if (near_unanimity && fn_arity != 3)
      cout << "near-unanimity ";
    
    if (conservative)
      cout << "conservative ";
    
    if (twosemilattice)
      cout << "2-semilattice ";
    
    if (maltsev)
      cout << "Mal'tsev ";
    
    cout << "operations.)\n";
  }
  else
    cout << " functions are not restricted to being of specific types.)\n";
}





// ***********************************************************************





// *************************************************************
// ***      Implementing the RelationEnumerator class        ***
// *************************************************************






RelationEnumerator::RelationEnumerator(nat arity, nat domain_size)
{
  this->arity = arity;
  this->domain_size = domain_size;
}




// ***********************




nat RelationEnumerator::generate_relations(bool quiet)
{
  // Deal with a couple of special cases first
  
  if (arity == 0)
  {
    if (!quiet)
      cout << "{}\n{<>}\n";
    return 2;
  }
  
  if (domain_size == 0 || domain_size == 1)
  {
    if (!quiet)
      cout << "{}\n";
    return 1;
  }

  // Now domain size is at least 2, and arity is at least 1
  nat sols;
  set<int> domain;
  for (nat i = 0; i < domain_size; ++i)
    domain.insert(i);

  // We need to create a RelationStore to find out the injective unary
  // polymorphisms of no relations over the domain we've created.  We
  // leave relstore empty.
  RelationStore relstore;
  relstore.set_real_domain(domain);

  ConstraintStore cs(arity, domain);
  cs.add_zero_one_range_constraint();
  cs.add_automorphism_constraint(relstore, true, true, false, false, true);

  sols=cs.find_polymorphisms(true, false, true, true, true,
			     quiet ? NULL : (SolutionReceiver::CallbackObject) this,
			     quiet ? NULL : (SolutionReceiver::CallbackFunction) &RelationEnumerator::polymorphism_receiver);
  
  return sols;
}




// *********************




void RelationEnumerator::polymorphism_receiver(const Function& f) const
{
  f.display_as_relation(cout);
}






// ***************************************************************************
// ***                       main program                                  ***
// ***************************************************************************

int main(int argc_i, char* argv[])
{
  nat argc = (nat) argc_i;                           // Coerce to be positive.
  try
  {
    RelationStore relstore;
    UserOptions opts;

    nat first_filename_pos = opts.read_options(argc, argv);

    // Handle the special case of being asked to generate relations,
    // rather than polymorphisms
    if (opts.get_enumerate_relations() >= 0)
    {
      if (first_filename_pos != argc)
      {
	cerr << "Error: Found unexpected filename: ";
	cerr << argv[first_filename_pos] << "\n";
	exit(10);
      }
      
      RelationEnumerator re(opts.get_fn_arity(),
			    opts.get_enumerate_relations());
      
      nat solutions = re.generate_relations(opts.get_quiet());
      
      cout << "\n" << solutions << " relations found.\n";
      exit(0);                                       // Normal exit
    }

    // Now we go on to the more usual case
    if (first_filename_pos == argc)
    {
      cout << "[No calculation requested]\n";
      exit(0);                                       // Normal exit
    }
    
    for (nat i = first_filename_pos; i < argc; i++)
      relstore.read_file(argv[i]);

    // Display the title
    cout << "\n" << relstore.get_title() << "\n";
    cout << string(min(relstore.get_title().length() + 1, 78), '=')
	 << "\n\n";
    opts.display_options_summary();
    
    // We assume that the domain we're working over is the union
    // of the domains present in the file.
    relstore.set_real_domain(relstore.get_apparent_domain());

    // Decompose relations
    if (opts.get_tractable() || opts.get_squash())
      relstore.squash_relations(false);
    
    relstore.decompose_all_relations();

    if (opts.get_tractable())
    {
      RelationStore::Tractability t = relstore.test_tractability();
      
      if (t == RelationStore::Unknown)
      {
	cout << "\nPolyanna cannot determine the tractability of the input relations.\n";
	exit(0);
      }
      if (t == RelationStore::Logspace)
      {
	cout << "\nThe input relations are solvable in LOGSPACE.\n";
	exit(1);
      }
      if (t == RelationStore::NLogspace)
      {
	cout << "\nThe input relations are solvable in NL.\n";
	exit(2);
      }
      if (t == RelationStore::PTime)
      {
	cout << "\nThe input relations are solvable in PTIME.\n";
	exit(3);
      }
      if (t == RelationStore::ConjecturedPTime)
      {
	cout << "\nThe input relations are conjectured to be solvable in PTIME.\n";
	exit(4);
      }
      if (t == RelationStore::NPComplete)
      {
	cout << "\nThe input relations are NP-complete.\n";
	exit(5);
      }
      cout << "Error: Unknown return code from RelationStore::test_tractability()\n";
      exit(10);
    }
      
    ConstraintStore cs(opts.get_fn_arity(), relstore.get_real_domain());

    // Add all requested constraints to the store
    if (opts.get_associative())
      cs.add_associative_constraint();
    if (opts.get_idempotent())
      cs.add_idempotent_constraint();
    if (opts.get_commutative())
      cs.add_commutative_constraint();
    if (opts.get_near_unanimity())
      cs.add_near_unanimity_constraint();
    if (opts.get_affine())
      cs.add_affine_constraint();
    if (opts.get_conservative())
      cs.add_conservative_constraint();
    if (opts.get_twosemilattice())
      cs.add_two_semilattice_constraint();
    if (opts.get_maltsev())
      cs.add_maltsev_constraint();

    if (opts.get_perm_opts() || opts.get_automorphisms())
      cs.add_automorphism_constraint(relstore,opts.get_perm_opts(),
				     false, false, opts.get_automorphisms(),
				     opts.get_force_all_symmetry_constraints());
    if (opts.get_essentially_nary())
      cs.add_essentiallynary_constraint();

    // We always need to add the polymorphism constraints
    relstore.add_polymorphism_constraints(cs);

    // Display, if necessary, the decomposed relations and the constraints
    // in the store
    if (opts.get_view_relations())
    {
      cout << "\nOn performing all possible decompositions, we get:\n\n";
      cout << relstore << "\n";
    }
    if (opts.get_view_constraints())
    {
      cout << "\nThe constraints in the store are:\n\n";
      cout << cs << "\n";
    }

    nat solutions = 0;
    for (nat i = 0; i < opts.get_iterations(); i++)
      solutions = cs.find_polymorphisms(opts.get_quiet(),
					opts.get_stop_at_first(),
					opts.get_force_complete_functions(),
					opts.get_force_all_symmetry_constraints(),
					opts.get_force_essential_constraints());
    
    if (opts.get_count() && opts.get_iterations() > 0)
      cout << "\nNumber of polymorphism specifications found: "
	   << solutions
	   << "\n";
  }

  // Catch every exception that may possibly be thrown, and which hasn't
  // definitely been caught elsewhere
  
  catch (UserOptions::UnknownOption o)
  {
    cerr << "Unknown option: " << o.option << ".\n";
    exit(10);
  }
  catch (UserOptions::MalformedOption o)
  {
    cerr << "Error parsing option "<< o.option << ".\n";
    exit(10);
  }
  catch (Relation::AritiesInconsistent)
  {
    cerr << "A relation has inconsistent arities.\n";
    exit(10);
  }
  catch (Relation::ReadError)
  {
    cerr << "An error occurred whilst reading a relation.\n";
    exit(10);
  }
  catch (RelationStore::ErrorOpeningFile e)
  {
    cerr << "An error occurred opening file " << e.filename << "\n";
    exit(10);
  }
  catch (RelationStore::ErrorReadingFile e)
  {
    cerr << "An error occurred reading file " << e.filename << "\n";
    exit(10);
  }
  catch (Relation::DecompositionError)
  {
    cerr << "One of the decomposition routines contains an internal bug.\n";
    exit(10);
  }
  catch (Relation::SanityError)
  {
    cerr << "An internal sanity check was violated in the Relation class.\n";
    exit(10);
  }
  catch (Relation::DomainError)
  {
    cerr << "An internal error has caused the apparent domain of a relation to be\n"
	 << "larger than its real domain.\n";
    exit(10);
  }
  catch (Relation::BelatedAlteration)
  {
    cerr << "An internal error has caused an attempt to be made to alter a relation\n"
	 << "after it has already been decomposed.\n";
    exit(10);
  }
  catch (ConstraintStore::SanityError)
  {
    cerr << "The ConstraintStore class has suffered an internal error.\n";
    exit(10);
  }
  catch (ConstraintStore::ConsistencyError)
  {
    cerr << "\n<Calculation aborted because of inconsistent demands.>\n";
    exit(10);
  }
  catch (ConstraintStore::SolutionExists sol)
  {
    cerr << "\n<Calculation aborted after first solution found.>\n";
    exit(10);
  }
  catch (Constraint::SanityError)
  {
    cerr << "A Constraint class has suffered an internal error.\n";
    exit(10);
  }
  catch (Function::SanityError)
  {
    cerr << "The Function class has suffered an internal error.\n";
    exit(10);
  }
  catch (FunctionValues::SanityError)
  {
    cerr << "The FunctionValues class has suffered an internal error.\n";
    exit(10);
  }
  catch (FunctionValues::EmptySet)
  {
    cerr << "The FunctionValues class has encountered an unexpected empty set.\n";
    exit(10);
  }
  catch (...)
  {
    cerr << "An unspecified error occurred.\n";
    exit(10);
  }
}
