// *************************************************************************
//
// relation_headers.h
//
// Part of Polyanna, Richard Gault, 2001.
//
// Copyright (C) University of Oxford, 2001, 2002
//
// *************************************************************************

// *************************************************************************
// ***                         CLASS DECLARATIONS                        ***
// *************************************************************************

#ifndef RELATION_HEADERS

#define RELATION_HEADERS

#include <iostream>
#include <vector>
#include <set>
#include <map>
#include <string>
#include "constraint_store_headers.h"

class Relation;
class RelationStore;


// *************************************************************************
// ***                       RELATIONSTORE CLASS                         ***
// *************************************************************************

// The RelationStore Class simply defines a repository in which relations
// may be placed.  More accurately, the repository contains pointers to
// relations, for reasons of speed.  Each repository has a title.  A
// RelationStore can automatically read in relations from a file.  If
// it is asked to read in relations from two files, it will take their
// union.  Provided the files contain titles, the repository's title
// will automatically be set up from these (though it may be user-modified
// if necessary).

class RelationStore : public SolutionReceiver
{
 public:

  // Constructor
  RelationStore();

  // Put a new relation into the store
  void add_relation(Relation* rp)
    {store.insert(rp);}

  // Check if specified pointer is in the store
  bool is_pointer_in_store(Relation* rp) const
    {return (store.find(rp) != store.end());}

  // Number of distinct pointers in the store
  nat get_size() const
    {return store.size();}

  // The apparent domain of the store.  This is somewhat slow as it is
  // calculated each time the function is called.
  std::set<int> get_apparent_domain() const;

  // The size of the apparent domain of the store.  Again, slightly slow.
  nat get_apparent_domain_size() const
    {return get_apparent_domain().size();}

  // Set the real domain (and propagate to relations in store)
  void set_real_domain(const std::set<int> &s);

  // Get the real domain.  (Fast.)
  std::set<int> get_real_domain() const
    {return real_domain;}

   // And find its size. (Fast.)
  nat get_real_domain_size() const
    {return real_domain.size();}

  // Read the indicated file into the store
  void read_file(const std::string filename);

  // Set the store's title
  void set_title(const std::string t)
    {title_given = true; title = t;}

  // Read the store's title.
  std::string get_title() const {return title;}

  // Extract a relation from the store.  (Or NULL if empty.)
  Relation* extract_relation();

  // Decompose all relations in the store
  bool decompose_all_relations();

  // Squash relations using the most powerful squashing fn.  Don't generate
  // any output if the quiet flag is true.
  bool squash_relations(bool quiet);


  // Force functions output to be polymorphisms
  void add_polymorphism_constraints(ConstraintStore& cs) const;

  
  // Finds IUPs.  Returns number found
  nat get_injective_unary_polymorphisms
    (std::set<std::map<int, int> >* mappings) const;


  // Different complexity classes
  enum Tractability
    {Logspace, NLogspace, PTime, NPComplete, ConjecturedPTime, Unknown};

  // Are the relations tractable?  Returns the lowest complexity class
  // that the relations live in (or Unknown).  If the quiet flag is set, then
  // no output is generated.
  Tractability test_tractability(bool quiet = false);

  // Show all relations in the store in tree-form
  void hierarchical_output(std::ostream& s) const;

  // Show all relations in the store in tuple-form
  void output_as_tuples(std::ostream& s) const;

  // ***************************************
  // *** RelationStore class: Exceptions ***
  // ***************************************

  // A SanityError is something which should never be thrown.
  class SanityError{};

  // File not found, typically.
  class ErrorOpeningFile
    {
    public:
      std::string filename;
      ErrorOpeningFile(std::string fn) {filename = fn;}
    };

  // The input file contains a syntax error.
  class ErrorReadingFile
    {
    public:
      std::string filename;
      ErrorReadingFile(std::string fn) {filename = fn;}
    };
  
 private:

  // The store of (pointers to) relations.
  std::set <Relation*> store;

  // Relation store iterator -- a handy abbreviation to save typing.
  typedef std::set<Relation*>::const_iterator RSI;

  // Title of the store.  Normally set automatically
  std::string title;

  // Have we got a title yet?
  bool title_given;

  // Max line length in file.  Kludge.  See implementation for comments about
  // this constant.
  static const nat max_line_length = 256;

  // The real domain.  (A superset of the apparent domain)
  std::set<int> real_domain;

  // The IUPs found, as a set of maps.
  mutable std::set<std::map<int,int> >* iup_maps;

  // Called when searching for IUPs
  void injective_polymorphism_receiver(const Function& f) const;

  // If squashing is on, contains the map used
  std::map<int, int> squash_map;

  // The size of the range of squash_map
  nat squash_range_size;

  // Number of domain elements unchanged under squashing
  nat squash_unchanged_size;

  // Called when searching for squashing fns
  void squashing_polymorphism_receiver(const Function& f);

  // Recursively put all children of relations in the store.  Returns
  // true if any added.
  bool flatten_store();

  // Just a helper function to the above, for internal use.
  void flatten_store_helper(std::set<Relation*> *rel_setp, Relation* rp);

  // Are the relations in store closed under a constant fn?
  // If quiet is true, then nop output will be generated on screen.
  bool test_for_constant(bool quiet = false);

  // A binary ACI function?
  bool test_for_aci(bool quiet = false);

  // A binary commutative conservative function?
  bool test_for_bcc(bool quiet = false);

  // An affine function?
  bool test_for_affine(bool quiet = false);

  // A Mal'tsev function?
  bool test_for_maltsev(bool quiet = false);

  // A near-unanimity function (of some arity)?
  bool test_for_nuf(bool quiet = false);

  // Essentially unary functions only?
  bool test_for_essentially_unary_only(bool quiet = false);

  // ***********************************
  //  Outputting a RelationStore object
  // ***********************************

  // Output on a stream
  friend std::ostream& operator<< (std::ostream& s, const RelationStore& rs)
    {
      rs.hierarchical_output(s);
      return s;
    }
};







// *************************************************************************
// ***                         RELATION CLASS                            ***
// *************************************************************************




// This class implements mathematical relations -- sets of tuples of domain
// elements.  In fact, this implementation restricts domain elements to
// being integers.

class Relation
{
 public:
  // Constructor: supply arity of relation
  Relation(nat a = 2);

  // Relation must be empty (or old arity==new)
  void change_arity(nat a);

  // Relation's arity
  nat get_arity() const
    {return arity;}

  // Relation's size
  nat get_size() const
    {return set_of_tuples.size();}

  // The apparent domain.  Recomputed on demand, so slow.
  std::set<int> get_apparent_domain() const;

   // The apparent domain's size.  Recomputed on demand, so slow.
  nat get_apparent_domain_size() const
    {return get_apparent_domain().size();}

  // Set the real domain (and propagate to any children)
  void set_real_domain(const std::set<int> &s);

  // Get the real domain
  std::set<int> get_real_domain() const
    {return real_domain;}

  // And find its size
  nat get_real_domain_size() const
    {return real_domain.size();}
  
  // Add tuple to relation.  Returns true if tuple is new.
  bool add_tuple(const std::vector <int> &t);

  // Remove tuple.  Returns true if successful
  bool remove_tuple(const std::vector <int> &t);
  
  // Make a copy of the relation, but without "side-info"
  Relation* make_top_level_copy();

  // Decompose relation (with children also decomposed)
  bool decompose();

  // Squash domain of the relation (inc. children)
  void squash_with_map(const std::map<int,int>& squash_map);
  
// Add constraints forcing fns to close this relation
  void construct_constraints(ConstraintStore& cs) const;
  
  // Return our children
  std::set<Relation*> get_children() const
    {return children;}

  // And the number of children we have
  nat number_of_children() const
    {return children.size();}

  // True iff the relation has no parent
  bool get_top_level()
    {return top_level;}

  // Could the relation be deleted?
  bool get_redundant() const
    {return redundant;}

  
  // Possible relation types.  Recall that AllowsAnything applies to
  // arity-1 relations only
  enum RelationType
    {Unknown, TrivialIrreducible, OtherIrreducible,
     DuplicatedColumns, CartesianPower, CartesianProduct, DisjointDomains,
     AllowsAnything, ConjunctiveProduct};

  // Gets what is known about the type of the relation
  RelationType get_type() const
    {return type;}

  // Displays relation and children as a tree
  void hierarchical_output(std::ostream& s) const;

  // Displays relation, but not children
  void output_as_tuples(std::ostream& s, bool new_lines = true) const;

  // ************************************
  // *** class Relation -- Exceptions ***
  // ************************************

  // Mix-up over arity of relation
  class AritiesInconsistent{};

  // A SanityError should never happen
  class SanityError{};

  // A DecompositionError is reallt just a special type of SanityError.
  // It should never happen
  class DecompositionError{};

  // Reading relation from file has gone wrong
  class ReadError{};

  // Real domain smaller than apparent domain
  class DomainError{};

  // Trying to alter an already-decomposed relation
  class BelatedAlteration{};


 private:
  // The arity of the relation
  nat arity;

  // The set of tuples which the relation is composed of.
  std::set < std::vector<int> > set_of_tuples;

  // Relation iterator.  Merely an abbrevisation to save typing
  typedef std::set< std::vector<int> >::const_iterator RI;

  // Always a superset of the apparent domain
  std::set<int> real_domain;

  // After decomposition, this is where we store our children
  std::set<Relation*> children;

  // Create a child with tuples etc. from copy_from, and put it in children.
  // NULL implies the child should be empty.
  Relation* make_child(Relation* copy_from);

  // Relation could be deleted without changing the set of its polymorphisms
  bool redundant;

  // The relation has no parent
  bool top_level;

  // What's known about the relation
  RelationType type;

  // We have had a squashing function applied
  bool squashed;

  // Type of decomposition function
  typedef bool (Relation::* DecompositionFunction) ();
  
  // Applys the given decomposition function
  bool apply_decomposition(DecompositionFunction decomposer);

  // Various decomposition functions for use by the above function.
  // They all check for the obvious decomposition implied by their name.
  bool check_for_trivial_relations();
  bool check_for_duplicated_columns();
  bool check_for_allows_anything();
  bool check_for_cartesian_power();
  bool check_for_cartesian_product();
  bool check_for_disjoint_domains();
  bool check_for_conjunctive_product();

  // **************************************************
  //  Helper functions for the decomposition functions
  // **************************************************

  // Delete columns for which keep[col] == false
  nat delete_columns(const std::vector<bool>& keep);

  // Given a putative split of a relation into two sub-relations, check
  // to see if the Cartesian product of the two relations does indeed yield
  // the original relation.
  bool check_split_for_cartesian_product(const std::vector<bool>& split);

  // ***********************************
  //        Outputting Relations
  // ***********************************

  // Output on a stream
  friend std::ostream& operator<< (std::ostream& s, const Relation& r)
    {
      r.output_as_tuples(s);
      return s;
    }
  
  // Input from a stream
  friend std::istream& operator>> (std::istream& s, Relation& r);
};

#endif
