
// ************************************************************************
//
// relation_store.cc.
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// ************************************************************************



// ************************************************************************
// ***     Defines the member functions of the RelationStore class      ***
// ************************************************************************


#include "relation_headers.h"
#include "misc_headers.h"
#include <set>
#include <iostream>
#include <fstream>

using namespace std;




// *********************************************************************




RelationStore::RelationStore()
{
  // Set the title to be some sensible default
  title = "Untitled set of relations";
  title_given = false;
}




// **********************************************************************




set <int> RelationStore::get_apparent_domain() const
{
  set<int> domain;
  set<int> relation_domain;
  
  for (RSI rel_pi = store.begin(); rel_pi != store.end(); rel_pi++)
  {
    relation_domain = (*rel_pi)->get_apparent_domain();
    for (set<int>::const_iterator ip = relation_domain.begin();
	 ip != relation_domain.end(); ip++)
      domain.insert(*ip);
  }
  
  return domain;
}




// **********************************************************************




void RelationStore::set_real_domain(const set<int> &s)
{
  real_domain = s;
  
  for (RSI rel_pi = store.begin(); rel_pi != store.end(); rel_pi++)
    (*rel_pi)->set_real_domain(s);         // Will throw a DomainError if the supplied domain is too small
}




// **********************************************************************




// Throughout this function, we need not worry about the file being left
// open after an exception, since it is automatically closed by its
// destructor on exit from the routine.

void RelationStore::read_file(const string filename)
{
  ifstream input(filename.c_str());
  if (!input)
    throw ErrorOpeningFile(filename);

  char c;
  string this_title = "";
  char line[max_line_length];
  bool first_line = true;
  
  while (input>>c, input.putback(c), c!='{' && !input.eof())
  {
    input.getline(line, max_line_length);
    
    // The following line is a workaround.  Something funny happens to
    // the input stream when input.getline hits an overlong line, and
    // doesn't reach its terminator.  An unremovable 0 seems to be
    // prepended to the stream.  This workaround simply throws an error
    // if an overlong line is read.  So lines of length > 255 aren't
    // supported.  Sorry!
    if (input.gcount() >= (int) max_line_length - 1)
      throw ErrorReadingFile(filename);

    if (first_line)
    {
      this_title = line;
    }
    else
    {
      this_title += "\n";
      this_title += line;
    }
    first_line = false;
  }
  
  if (this_title == "")
    this_title = "Untitled relations from file " + filename;
  
  if (title_given)
    title += "\nand " + this_title;
  else
    title = this_title;
  
  title_given = true;

  Relation* rel_ptr = NULL;
  try
  {
    while (input >> c, input.putback(c), !input.eof())
    {
      rel_ptr = new(Relation);
      if (c != '{')
	throw ErrorReadingFile(filename);
      
      input >> (*rel_ptr);
      add_relation(rel_ptr);
      rel_ptr = NULL;
    }
  }
  
  catch(...)
  {
    // Free memory reserved for the faulty relation
    if (rel_ptr != NULL)
      delete(rel_ptr);
    
    throw ErrorReadingFile(filename);
  }
}




// **********************************************************************




Relation* RelationStore::extract_relation()
{
  if (store.size() == 0)
    return NULL;

  RSI sp = store.begin();
  Relation* rp = (*sp);
  nat n = store.erase(rp);

  if (n != 1)                   // Duplicated value in set!
    throw SanityError();
  
  return rp;
}




// **********************************************************************




bool RelationStore::decompose_all_relations()
{
  bool at_least_one_decomposition = false;

  for (RSI rpi = store.begin(); rpi != store.end(); ++rpi)
    if ((*rpi)->decompose())
      at_least_one_decomposition = true;

  if (at_least_one_decomposition)
    if (!flatten_store())
      throw SanityError();    // Done decomposition, but not come up with any new relations?!?

  return at_least_one_decomposition;
}




// **********************************************************************




// Depth-first search

bool RelationStore::flatten_store()
{
  Relation* rp;
  set<Relation*> target_set;

  while ((rp = extract_relation()) != NULL)
  {
    target_set.insert(rp);
    flatten_store_helper(&target_set, rp);
  }
  store.swap(target_set);

  if (target_set.size() == store.size())
    return false;
  
  return true;
}


// ------


// This helper function actually does the work of the depth-first search

void RelationStore::flatten_store_helper(set<Relation*> *rel_setp,
					 Relation* rp)
{
  set <Relation*> children = rp->get_children();
  
  for (set<Relation*>::const_iterator rpi = children.begin();
       rpi != children.end();
       ++rpi)
  {
    rel_setp->insert(*rpi);
    flatten_store_helper(rel_setp, *rpi);
  }
}




// **********************************************************************




void RelationStore::add_polymorphism_constraints(ConstraintStore& cs) const
{
  // No need to add constraints for redundant relations
  for (RSI rpi = store.begin(); rpi != store.end(); ++rpi)
    if (!((*rpi)->get_redundant()))
      (*rpi)->construct_constraints(cs);
}




// ***********************************************************************




nat RelationStore::get_injective_unary_polymorphisms(set<map<int, int> >* mappings) const
{
  if (mappings->size() != 0)
    throw SanityError();

  iup_maps = mappings;

  // Set up a constraint store to search for arity-1, completely specified,
  // squashing polymorphisms
  ConstraintStore cs(1, real_domain);
  add_polymorphism_constraints(cs);

  // Find the polymorphisms (and use injective_polymorphism_receiver to
  // update the set of mappings)
  cs.find_polymorphisms(true, false, true, true, true,
			(SolutionReceiver::CallbackObject) this,
			(SolutionReceiver::CallbackFunction) &RelationStore::injective_polymorphism_receiver);
  
  return iup_maps->size();
}




// ***********************************************************************




void RelationStore::injective_polymorphism_receiver(const Function& f) const
{
  // Determine the range of f, and copy f into a map
  set<int> f_range;
  f_range.clear();
  map <int,int> f_map;

  for (set<int>::const_iterator isi = real_domain.begin();
       isi != real_domain.end();
       ++isi)
  {
    vector<int> v(1);
    v[0] = *isi;
    const FunctionArguments* v_fa = FunctionArguments::tuple_to_function_arguments(v);
    int fi = (f[v_fa])->min_value();                       // f(*isi)
    f_range.insert(fi);
    f_map[*isi] = fi;
  }

  // Is f injective?
  // If so, Add f_map to the set of injective unary polymorphisms we have
  // found so far.  This uses copying, but this is no big deal: how big are
  // our domains going to get?
  if (f_range.size() == real_domain.size())
    iup_maps->insert(f_map);

  return;
}




// ***********************************************************************




bool RelationStore::squash_relations(bool quiet)
{
  // Initialise variables needed for the squashing calculation
  squash_map.clear();
  
  for (set<int>::const_iterator isi = real_domain.begin();
       isi != real_domain.end();
       ++isi)
    squash_map[*isi] = *isi;
  
  squash_range_size = real_domain.size();
  squash_unchanged_size = real_domain.size();

  // Set up a constraint store to search for arity-1, completely specified,
  // squashing polymorphisms
  ConstraintStore cs(1, real_domain);
  add_polymorphism_constraints(cs);
  cs.add_squashing_constraint();

  // Find the polymorphisms (and use squashing_polymorphism_receiver to
  // update the squash_* variables)
  cs.find_polymorphisms(true, false, true, true, true,
			(SolutionReceiver::CallbackObject) this,
			(SolutionReceiver::CallbackFunction) &RelationStore::squashing_polymorphism_receiver);
  
  if (squash_range_size == real_domain.size())
  {
    if (!quiet)
      cout << "\n[No squashing functions can be applied.]\n\n";
    return false;
  }

  // Even if the quiet flag is set, we can't really keep this information
  // from the user
  cout << "\n[Applying the squashing function: ";
  for (set<int>::const_iterator isi = real_domain.begin();
       isi != real_domain.end();
       ++isi)
  {
    if (isi != real_domain.begin())
      cout << "; ";
    
    cout << *isi << "->" << squash_map[*isi];
  }
  cout << ".]\n\n";

  
  // Now actually squash the individual relations
  for (RSI rpi = store.begin(); rpi != store.end(); ++rpi)
    if ((*rpi)->get_top_level())
      (*rpi)->squash_with_map(squash_map);

  // Our real domain has now shrunk.  Find out what it is.
  set<int> squash_range;
  squash_range.clear();
  
  for (map<int,int>::const_iterator mi = squash_map.begin();
       mi != squash_map.end();
       ++mi)
    squash_range.insert(mi->second);
  
  set_real_domain(squash_range);
  
  return true;
}



// ***********************************************************************




void RelationStore::squashing_polymorphism_receiver(const Function& f)
{
  // Determine the range of f, and the number of elements mapped to themselves
  set<int> f_range;
  f_range.clear();
  nat f_unchanged_size = 0;
  
  for (set<int>::const_iterator isi = real_domain.begin();
       isi != real_domain.end();
       ++isi)
  {
    vector<int> v(1);
    v[0] = *isi;
    const FunctionArguments* v_fa = FunctionArguments::tuple_to_function_arguments(v);
    int fi = (f[v_fa])->min_value();                          // f(*isi)
    if (fi == *isi)
      f_unchanged_size++;
    
    f_range.insert(fi);
  }

  if (f_range.size() < squash_range_size ||
      (f_range.size() == squash_range_size &&
       f_unchanged_size > squash_unchanged_size)
      )
  {
    // f induces a better squash_map than we have so far.  Update
    // squash_* variables appropriately.
    // This, rather wastefully, is almost verbatim the same as the last
    // loop.  But it's not really worth doing anything very clever here.
    // How big are our domains going to get?
    for (set<int>::const_iterator isi = real_domain.begin();
	 isi != real_domain.end();
	 ++isi)
    {
      vector<int> v(1);
      v[0] = *isi;
      const FunctionArguments* v_fa = FunctionArguments::tuple_to_function_arguments(v);
      int fi = (f[v_fa])->min_value();                               // f(*isi)
      squash_map[*isi] = fi;
    }
    
    squash_range_size = f_range.size();
    squash_unchanged_size = f_unchanged_size;
  }
}




// ***********************************************************************




// Test for constant functions

bool RelationStore::test_for_constant(bool quiet)
{
  cout << "Checking for constant functions..."
       << flush;
  
  ConstraintStore cs_const(0, real_domain);
  add_polymorphism_constraints(cs_const);
  
  if (cs_const.find_polymorphisms(quiet, false, false, false, false) > 0)
  {
    if (quiet)
      cout << "yes.\n";
    return true;
  }
  else
  {
    cout << "No.\n";
    return false;
  }
}




// ***********************************************************************




// Test for semilattice polymorphisms

bool RelationStore::test_for_aci(bool quiet)
{
  try
  {
    cout << "Checking for semilattice functions..."
	 << flush;
    
    ConstraintStore cs_aci(2, real_domain);
    add_polymorphism_constraints(cs_aci);
   
    cs_aci.add_associative_constraint();
    cs_aci.add_idempotent_constraint();
    cs_aci.add_commutative_constraint();
    
    cs_aci.find_polymorphisms(quiet, true, false, false, false);
    cout << "No.\n";
    return false;
  }
  
  catch(ConstraintStore::SolutionExists& sol)
  {
    if (quiet)
      cout << "Yes.\n";
    
    return true;
  }
}




// ***********************************************************************




// Test for binary conservative commutative polymorphisms

bool RelationStore::test_for_bcc(bool quiet)
{
  try
  {
    cout << "Checking for binary, commutative, conservative functions..."
	 << flush;
    
    ConstraintStore cs_bct(2, real_domain);
    add_polymorphism_constraints(cs_bct);
    
    cs_bct.add_conservative_constraint();
    cs_bct.add_commutative_constraint();
    
    cs_bct.find_polymorphisms(quiet, true, false, false, false);
    
    cout << "No.\n";
    return false;
  }
  
  catch(ConstraintStore::SolutionExists& sol)
  {
    if (quiet)
      cout << "Yes.\n";
    
    return true;
  }
}




// ***********************************************************************




// Test for affine functions

bool RelationStore::test_for_affine(bool quiet)
{
  try
  {
    cout << "Checking for affine functions..."
	 << flush;
    
    ConstraintStore cs_affine(3, real_domain);
    add_polymorphism_constraints(cs_affine);
    
    cs_affine.add_affine_constraint();
    
    cs_affine.find_polymorphisms(quiet, true, false, false, false);
    
    cout << "No.\n";
    return false;
  }
  
  catch(ConstraintStore::SolutionExists& sol)
  {
    if (quiet)
      cout << "Yes.\n";
    
    return true;
  }
}





// ***********************************************************************




// Test for Mal'tsev functions

bool RelationStore::test_for_maltsev(bool quiet)
{
  try
  {
    cout << "Checking for Mal'tsev functions..."
	 << flush;
    
    ConstraintStore cs_maltsev(3, real_domain);
    add_polymorphism_constraints(cs_maltsev);
    
    cs_maltsev.add_maltsev_constraint();
    
    cs_maltsev.find_polymorphisms(quiet, true, false, false, false);
    
    cout << "No.\n";
    return false;
  }
  
  catch(ConstraintStore::SolutionExists& sol)
  {
    if (quiet)
      cout << "Yes.\n";
    
    return true;
  }
}




// ***********************************************************************



// Test for NUF functions

bool RelationStore::test_for_nuf(bool quiet)
{
  try
  {
    nat upper_bound = real_domain.size();
    if (upper_bound < 3)
      upper_bound = 3;
    
    for (nat arity = 3; arity <= upper_bound; ++arity)
    {
      if (arity == 3)
	cout << "Checking for majority functions..."
	     << flush;
      else
	cout << "Checking for near-unanimity functions of arity "
	     << arity << "..."
	     << flush;
      
      ConstraintStore cs_nuf(arity, real_domain);
      add_polymorphism_constraints(cs_nuf);
      
      cs_nuf.add_near_unanimity_constraint();
      
      cs_nuf.find_polymorphisms(quiet, true, false, false, false);
      cout << "No\n";
    }
    return false;
  }
  
  catch(ConstraintStore::SolutionExists& sol)
  {
    if (quiet)
      cout << "Yes.\n";
    
    return true;
  }
}





// ***********************************************************************




// Test for essentially unary functions only

bool RelationStore::test_for_essentially_unary_only(bool quiet)
{
  try
  {
    nat upper_bound = real_domain.size();
    if (upper_bound < 3)
      upper_bound = 3;
    
    for (nat arity = 2; arity <= upper_bound; ++arity)
    {
      cout << "Checking for any functions of essential arity "
	   << arity << "..."
	   << flush;
      
      ConstraintStore cs_npc(arity, real_domain);
      add_polymorphism_constraints(cs_npc);
      
      cs_npc.add_essentiallynary_constraint();
      
      // The aqutomorphism constraint makes little difference
      cs_npc.add_automorphism_constraint(*this, true, false, false,
					 false, false);
      cs_npc.add_idempotent_constraint();
      
      cs_npc.find_polymorphisms(quiet, true, false, true, true);
      cout << "No.\n";
    }
    return true;     // Backwards logic compared with other tests!
  }
  
  catch(ConstraintStore::SolutionExists& sol)
  {
    if (quiet)
      cout << "Yes.\n";
    
    return false;
  }
}




// ***********************************************************************




RelationStore::Tractability RelationStore::test_tractability(bool quiet)
{
  // Double-check that we've already applied a squashing function
  if (squash_relations(true))
  {
    cerr << "Warning: Relations have been squashed after being decomposed.\n";
    cerr << "         This may indicate a programming error.\n";
    exit(10);
  }

  // So far, we know nothing of the relations' tractability
  Tractability tractability = Unknown;
  
  // If there's only one top-level relation, we might be able to split it
  // up via conjunctive products or disjoint domains, and analyse the
  // subrelations separately.
  // We do this (if possible) before analysing the relation directly,
  // since it speeds things up dramatically if such a decomposition can
  // be performed.

  // Find out how many top-level non-redundant relations there are
  nat top_level = 0;
  Relation* top_level_rln = NULL;     // Initialised to stop -Wall complaining
  set<Relation*> rel_set;
  bool decomposition_possible = false;
  
  for (RSI rsi = store.begin(); rsi != store.end(); ++rsi)
  {
    if (!(*rsi)->get_redundant() && (*rsi)->get_top_level())
    {
      top_level_rln = *rsi;
      top_level++;
    }
  }

  if (top_level == 1 &&
      top_level_rln->get_type() == Relation::DisjointDomains)
  {
    cout << "\nThe input relation reduces to the disjoint union of two or more relations.\n";
    cout << "It is tractable if all component relations are tractable.\n";
    rel_set = top_level_rln->get_children();
    decomposition_possible = true;
  }
  else if (top_level == 1 &&
	   top_level_rln->get_type() == Relation::ConjunctiveProduct)
  {
    cout << "\nThe input relation reduces to the conjunctive product of two or more\n";
    cout << "relations.  It is tractable if all component relations are tractable.\n";
    rel_set = top_level_rln->get_children();
    decomposition_possible = true;
  }

  if (decomposition_possible)
  {
    // If all the sub-relations are tractable, then so is the full relation.

    // The relation is going to be (at best) PTime-solvable
    Tractability sub_tractability = PTime;
    nat relation_number = 0;
    
    for (RSI rsi = rel_set.begin(); rsi != rel_set.end(); ++rsi)
    {
      cout << "\nTesting relation " << ++relation_number
	   << " out of " << rel_set.size() << ":\n";
      RelationStore relstore;
      
      // We want to put the sole relation (**rsi) into relstore.  But we
      // can't just put the pointer in with  relstore.add_relation because
      //   (a) the relation will not be top-level, and
      //   (b) relstore might want to change the relation by applying
      //       squashing functions to it.
      // So we must make a copy of (**rsi), and insert a pointer to that
      // relation into relstore.
      Relation* relation = (*rsi)->make_top_level_copy();
      relstore.add_relation(relation);
      relstore.set_real_domain(relation->get_apparent_domain());
      relstore.squash_relations(true);
      relstore.decompose_all_relations();
      
      Tractability t = relstore.test_tractability(quiet);
      
      if (t == Unknown || t == NPComplete)
      {
	// We've not learned anything from splitting
	sub_tractability = Unknown;
	break;
      }
      else if (t == ConjecturedPTime)
      {
	// Here, we also know that sub_tractability == PTime or ConjPTime
	sub_tractability = ConjecturedPTime;
      }
      else if (t == PTime && sub_tractability == PTime)
      {
	sub_tractability = PTime;
      }
      else
      {
	cerr << "Warning: Programming error.  A sub-relation has returned an unknown tractability\n";
	cerr << "         value.\n";
      }
      
      delete(relation);
    }
    
    // Every sub-relation must have been PTime or ConjecturedPTime.
    // sub_tractability is ConjecturedPTime iff at least one of the
    // subrelations was ConjecturedPTime (else it's PTime).
    if (sub_tractability == PTime)
      return sub_tractability;
    
    if (sub_tractability == ConjecturedPTime)
      tractability = sub_tractability;
  }  // if (decomposition_possible)

  // At this point, tractability is either Unknown or, less likely,
  // ConjecturedPTime.  This latter indicates that the input relations could
  // be decomposed into disjoint domains or conjunctive products, and
  // that each sub-relation was either tractable or conjectured tractable
  // (with at least one sub-relation being conjectured tractable).
  
  // We now see if we can determine the tractability of the relations in
  // the store be analysing polymorphisms directly.
  
  if (test_for_constant(quiet))
    return Logspace;
  
  if (test_for_aci(quiet))
    return PTime;
  
  if (test_for_bcc(quiet))
    return PTime;
  
  if (test_for_affine(quiet))
    return PTime;
  
  if (test_for_nuf(quiet))
    return PTime;

  // For the case of Mal'tsev, don't return the tractability directly
  // (since it is only conjectured).  Instead, see if we can find out more.
  if (test_for_maltsev(quiet))
    tractability = ConjecturedPTime;
  
  if (test_for_essentially_unary_only(quiet))
  {
    // This indicates NP-completeness.
    if (tractability == ConjecturedPTime)
    {
      cerr << "Warning: Relations are both conjectured PTime and NP-complete.\n";
      cerr << "         Even if the conjecture is wrong, this should not happen.\n";
      cerr << "         A programming error has occurred.\n";
    }
    return NPComplete;
  }
  
  // At this stage, tractability will be either Unknown or ConjecturedPTime
  return tractability;
}




// ***********************************************************************




void RelationStore::hierarchical_output(ostream& s) const
{
  for (RSI rpi = store.begin(); rpi != store.end(); ++rpi)
  {
    if ((*rpi)->get_top_level())
    {
      s << **rpi << " == ";
      (*rpi)->hierarchical_output(s);
      s << "\n\n";
    }
  }
}




// ***********************************************************************




void RelationStore::output_as_tuples(ostream& s) const
{
  for(RSI rel_pi = store.begin(); rel_pi != store.end(); ++rel_pi)
    s << (**rel_pi) << "\n";
}
