// ***********************************************************************
//
// solver.cc
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// ***********************************************************************

#include "constraint_store_headers.h"
#include "misc_headers.h"

using namespace std;

// This file contains the routines which actually solve the constraint
// problem.  The public entry-point to these routines is
// ConstraintStore::find_polymorphims()

nat ConstraintStore::find_polymorphisms
                     (bool quiet,
		      bool stop_at_first,
		      bool force_complete_functions,
		      bool force_all_symmetry_constraints,
		      bool force_essential_constraints,
		      SolutionReceiver::CallbackObject co,
		      SolutionReceiver::CallbackFunction cf)
{
  // Do a final "pre-launch" check
  check_store_consistency();

  // The rest of this function is devoted to initialising the "global"
  // variables that the solver needs.

  // options is a handy structure used for carrying around options used
  // whilst solving.
  SolverOptions options;
  options.quiet = quiet;
  options.stop_at_first = stop_at_first;
  options.force_complete_functions = force_complete_functions;
  options.force_all_symmetry_constraints = force_all_symmetry_constraints;
  options.force_essential_constraints = force_essential_constraints;

  callback_object = co;
  callback_function = cf;

  // f is the function we'll apply constraints to
  Function f(arity, domain, lex_ordering);

  // max_depth is the upper limit on search depth.  +1 for 'complete' search
  max_depth = constrained_fas.size() + 1;

  // Copy set constraints into list unsatisfied_constraints: no constraint
  // is yet satisfied (as far as we know)
  copy(constraints.begin(),
       constraints.end(),
       back_inserter(unsatisfied_constraints));
  satisfied_constraints.resize(max_depth + 1);
  
  // Check that satisfied_constraints[] is empty
  for (nat i = 0; i < max_depth + 1; ++i)
  {
    if (satisfied_constraints[i].size() != 0)
      throw SanityError();
  }

  // Copy the set constrained_fas into the list redundant_fas: no
  // FunctionArgument is yet redundant (AFAWK)
  copy(constrained_fas.begin(),
       constrained_fas.end(),
       back_inserter(unfinished_fas));
  redundant_fas.resize(max_depth + 1);
  
  // Check that redundant_fas[] is empty
  for (nat i = 0; i < max_depth + 1; ++i)
  {
    if (redundant_fas[i].size() != 0)
      throw SanityError();
  }

  // Set up the lookup table lookup_singleton_function_values.
  // We can only do this if the domain supplied by the user is sensible --
  // defined in the following line to ba subset of {0, 1, ..., 999}
  if (domain.size() > 0 &&
      min_set_value(domain) >= 0 &&
      max_set_value(domain) < 1000) // Range is arbitrary
  {
    lookup_singleton_function_values.clear();
    lookup_singleton_function_values.resize(max_set_value(domain) + 1);
    
    for (set<int>::const_iterator si = domain.begin();
	 si != domain.end();
	 ++si)
    {
      set<int> s;
      s.insert(*si);
      lookup_singleton_function_values[*si] =
	FunctionValues::set_to_function_values(s);
    }
    
    lookup_sfv_available = true;
  }
  else
  {
    lookup_sfv_available = false;
    if (domain.size() != 0)
    {
      cerr << "Warning: Your use of very large, or negative domain values\n";
      cerr << "         may slow the calculation down slightly.\n";
    }
  }

  depth = 0;
  solution_count = 0;

  // Initialise any notifiable constraints
  for (CSI csi = notifiable_constraints.begin();
       csi != notifiable_constraints.end();
       ++csi)
  {
    (*csi)->notify_search_starting(max_depth);
  }

  // Actually carry out the search
  set <const FunctionArguments*> modified_fas = constrained_fas;
  try
  {
    do_search(f, options, modified_fas, false);
  }
  catch(...)
  {
    tidy_up();
    throw;
  }
  
  tidy_up();
  return solution_count;
}



// ******************



void ConstraintStore::tidy_up()
{
  // Tidy up: remove all FAs from unfinished_fas, and all constraints
  // from unsatisfied_constraints
  unfinished_fas.clear();
  unsatisfied_constraints.clear();

  // Allow any notifiable constraints to tidy up
  for (CSI csi = notifiable_constraints.begin();
       csi != notifiable_constraints.end();
       ++csi)
  {
    (*csi)->notify_search_ending();
  }
}




// ****************




void ConstraintStore::do_search
                      (Function& f,
		       const SolverOptions& options,
		       set<const FunctionArguments*>& modified_fas,
		       bool complete)
{
  // Tell any unsatisfied notifiable constraints that we've moved down a level
  for (CLI cli = unsatisfied_constraints.begin();
       cli != unsatisfied_constraints.end();
       ++cli)
  {
    if ((*cli)->need_level_notification())
      (*cli)->notify_search_descent();
  }

  f.save_state();
 
  Action action = apply_constraints(f, options, modified_fas, complete);

  if (action == Backtrack)
  {
    do_backtrack(f);
    return;
  }

  if (action == FoundSolution)
  {
    // Solution found
    solution_count++;
    
    if (callback_object != NULL)
      (callback_object->*callback_function)(f);

    if (!options.quiet)
      cout << "\n-------- " << solution_count << "-------\n\n" << f;
    else
      if (solution_count % 10000 == 0) cerr << "." << flush;
    
    if (options.stop_at_first)
      throw SolutionExists(f);
    
    do_backtrack(f);

    return;
  }

  if (complete)
  {
    // No point going any further
    do_backtrack(f);
    return;
  }

  // Find the FunctionArgument we're going to assign to next: the FA which
  // is most constrained.  That is, we use the fail-first heuristic.
  
  list<const FunctionArguments*>::iterator mc_fali;   // Most constrained FunctionArgument iterator ...
  nat mc_fa_size = domain.size() + 42;                // and the number of values it can take (42 is arbitrary!)
  nat assignable_fas = 0;                             // Number of unfinished_fas members which can take >=2 values

  // This is really a for loop, but it can't be, as we're deleting list
  // members as we iterate over them.
  list<const FunctionArguments*>::iterator fali = unfinished_fas.begin();
  while (fali != unfinished_fas.end())
  {
    const FunctionArguments* this_fa = *fali;
    nat this_fa_size = f[this_fa]->get_size();

    if (this_fa_size == 1)
    {
      // It's safe to delete (*fali) from unfinished_fas whilst iterating
      // over it, because it's a list.
      // next_fali is a circumlocution to avoid us following an out-of-date
      // iterator
      list<const FunctionArguments*>::iterator next_fali = ++fali;
      
      redundant_fas[depth].splice(redundant_fas[depth].begin(),
				  unfinished_fas,
				  --fali);
      fali = next_fali;
      
      continue;
    }

    if (this_fa_size == 0)
    {
      // Constraint violated.  Should've really spotted this before
      do_backtrack(f);
      return;
    }

    assignable_fas++;
    if (this_fa_size < mc_fa_size)
    {
      // This is potentially a new most-constrained FunctionArguments.  But
      // we've first got to check that some constraint constrains it
      // (unless we're looking for complete functions only, of course).
      // Two cases: either we want to break symmetry constraints, or we do
      // not (if we can sensibly avoid it).
      if (!options.force_complete_functions)
      {
	bool constrained = false;
	if (options.force_all_symmetry_constraints)
	{
	  // First case:
	  for (CLI cli = unsatisfied_constraints.begin();
	       cli != unsatisfied_constraints.end();
	       ++cli)
	  {
	    if ((*cli)->is_constraint_relevant(this_fa) >=
		                           Constraint::SymmetryConstrained)
	    {
	      constrained = true;
	      break;
	    }
	  }
	}
	else
	{
	  // Second case:
	  for (CLI cli = unsatisfied_constraints.begin();
	       cli != unsatisfied_constraints.end();
	       ++cli)
	  {
	     if ((*cli)->is_constraint_relevant(this_fa) ==
		                                      Constraint::Constrained)
	     {
	       constrained = true;
	       break;
	     }
	  }
	}
	// Both cases: This FA is not constrained by any constraint --
	// mark it as redundant.
	if (!constrained)
	{
	  list<const FunctionArguments*>::iterator next_fali = ++fali;
	  redundant_fas[depth].splice(redundant_fas[depth].begin(),
				      unfinished_fas,
				      --fali);
	  fali = next_fali;
	  assignable_fas--;             // We can't be assigned to, after all.
	  continue;
	}
      }
      // Finally: we know that we really have found a new most-constrained
      // FunctionArguments tuple.
      mc_fa_size = this_fa_size;
      mc_fali = fali;
    }

    if (mc_fa_size == 2 && assignable_fas >= 2)
    {
      // This is as constrained an FA as we're going to get.
      // (The assignable_fas check is needed so we know if f is complete)
      break;
    }
    
    ++fali;
  }
  
  // Now, all unfinished FAs must have size 1: ie. the function is complete
  // wrt the constrained FAs
  if (assignable_fas == 0)
  {
    modified_fas.clear();
    depth++;
    do_search(f, options, modified_fas, true);
    do_backtrack(f);
    return;
  }

  // We have found a FunctionArgument which is constrained, and which can
  // take at least two values
  // First, save the FunctionArgument we have found
  const FunctionArguments* mc_fa = *mc_fali;

  // Next, make the FunctionArgument redundant, so it won't be looked at
  // when we recurse.
  // (This would happen next time round the loop in any case, but it's
  // faster here.)
  redundant_fas[depth].splice(redundant_fas[depth].begin(),
			      unfinished_fas,
			      mc_fali);

  // Is there only one FunctionArgument needing assignment?  If so, f will
  // be complete when we recurse
  complete = (assignable_fas == 1);

  // Get the values which f(mc_fa) can take
  const set<int> mc_values = f[mc_fa]->function_values_to_set();

  // If lookup is availiable, then we get a speed increase
  if (lookup_sfv_available)
  {
    for (set<int>::const_iterator si = mc_values.begin();
	 si != mc_values.end();
	 ++si)
    {
      f.set_value(mc_fa, lookup_singleton_function_values[*si]);
      // Only one FunctionArguments tuple has been modified so far: mc_fa.
      // Need to set up modified_fas each time round the loop, as do_search
      // will corrupt it.
      modified_fas.clear();
      modified_fas.insert(mc_fa);
      
      // Recurse.
      depth++;
      do_search(f, options, modified_fas, complete);
    }
  }
  else   // Lookup not available
  {
    for (set<int>::const_iterator si = mc_values.begin();
	 si != mc_values.end();
	 ++si)
    {
      // Turn each singleton value into a FunctionValues argument.
      set<int> s;
      s.insert(*si);
      f.set_value(mc_fa, FunctionValues::set_to_function_values(s));
      
      // Only one FunctionArguments tuple has been modified so far: mc_fa.
      // Need to set up modified_fas each time round the loop, as do_search
      // will corrupt it.
      modified_fas.clear();
      modified_fas.insert(mc_fa);
      
      // Recurse.
      depth++;
      do_search(f, options, modified_fas, complete);
    }
  }
  
  do_backtrack(f);
  return;
}




// *************




inline void ConstraintStore::do_backtrack(Function& f)
{
  // Put back any constraints which we removed through being AlwaysSatisfied...
  unsatisfied_constraints.splice(unsatisfied_constraints.begin(),
				 satisfied_constraints[depth]);
  
  // ...and any FunctionArguments which we removed through being redundant...
  unfinished_fas.splice(unfinished_fas.begin(),
			redundant_fas[depth]);

  // Tell any unsatisfied notifiable constraints that we're backtracking
  for (CLI cli = unsatisfied_constraints.begin();
       cli != unsatisfied_constraints.end();
       ++cli)
  {
    if ((*cli)->need_level_notification())
      (*cli)->notify_search_ascent();
  }
  
  // Restore f
  f.restore_state();

  depth--;
  
  return;
}




// ****************




inline ConstraintStore::Action ConstraintStore::apply_constraints
                                  (Function& f,
				   const SolverOptions& options,
				   set<const FunctionArguments*>& modified_fas,
				   bool complete)
{
  // The following variables hold the number of constraints whose
  // satisfiability level is Unknown, SymmetryBreaking, etc.
  // We don't need a never_sat_num, because as soon as we spot a single
  // NeverSatisfied constraint, we return Backtrack.
  // Also, we don't need an always_sat_num.  There's no obvious reason why
  // we don't: we just don't happen to need it.
  nat unknown_num;
  nat symmetry_break_num;
  nat sometimes_sat_num;
  
  // Actually apply all the constraints, and keep track of the number of
  // different satisfaction levels.
  // To do this, repeatedly apply constraints until they stop modifying the
  // function.  This is a simpler version than the one we had originally,
  // and seems to run faster.  I would have expected the old version to be
  // the faster, but there you go...
  do{
    unknown_num = 0;
    symmetry_break_num = 0;
    sometimes_sat_num = 0;
    set<const FunctionArguments*> cumulative_new_mod_fas;
    
    // This is really a for loop, but can't be because we delete iterators
    // we're iterating over
    CLI cli = unsatisfied_constraints.begin();
    while (cli != unsatisfied_constraints.end())
    {
      set<const FunctionArguments*> new_modified_fas;

      Constraint::SatLevel sat_level =
	(*cli)->apply_constraint(f, modified_fas, new_modified_fas, complete);

      ++cli;
      
      // Note that from here on down, cli might be
      // unsatisfied_constraints.end(), so we can't dereference it directly

      switch (sat_level)
      {
      case Constraint::NeverSatisfied:
	return Backtrack;
	
      case Constraint::AlwaysSatisfied:
	if (!complete)
	{
	  // Remove the constraint from the list of unsatisfied constraints
	  // A circumlocution to make sure that cli remains a valid iterator
	  list<Constraint*>::iterator next_cli = cli;
	  satisfied_constraints[depth].splice(satisfied_constraints[depth].begin(),
					      unsatisfied_constraints,
					      --cli);
	  cli = next_cli;
	}
	break;
	
      case Constraint::SometimesSatisfied:
	++sometimes_sat_num;
	break;
	
      case Constraint::Unknown:
	++unknown_num;
	break;
	
      case Constraint::SymmetryBreaking:
	++symmetry_break_num;
      }

      // Union modifed_fas and cumulative_new_mod_fas with the newly modified
      // FunctionArguments.  Doing this by hand seems to be faster than using
      // the built-in union function -- perhaps because we only need to
      // execute the loop once.
      for (set<const FunctionArguments*>::const_iterator fasi =
	     new_modified_fas.begin();
	   fasi != new_modified_fas.end();
	   ++fasi)
      {
	modified_fas.insert(*fasi);
	cumulative_new_mod_fas.insert(*fasi);
      }
    }  // end while
    
    modified_fas.swap(cumulative_new_mod_fas);
    
  } while (modified_fas.size() != 0);

  // In this section, we assume that any never-satsified constraint has
  // already been spotted as such (i.e., within the constraint itself), and
  // so we will have Backtracked above.  If it hasn't been, the solver will
  // notice later (if complete is false), or the function output
  // routine will print a warning, reminding us to fix the buggy constraint.

  // ***TODO*** This last is not yet implemented.
  
  if (complete)
    return FoundSolution;
  
  if (options.force_complete_functions)
    return SearchOn;
  
  if (unknown_num > 0)
    return SearchOn;
  
  if (options.force_all_symmetry_constraints && symmetry_break_num > 0)
    return SearchOn;
  
  if (options.force_essential_constraints && sometimes_sat_num > 0)
    return SearchOn;
  
  if (sometimes_sat_num <= 1)
    return FoundSolution;
  
  return SearchOn;
}




// ****************
