// **********************************************************************
//
// symmetry.cc
//
// Part of Polyanna, Richard Gault, 2001
//
// Copyright (C) University of Oxford, 2001, 2002
//
// **********************************************************************

#include "constraints_headers.h"
#include "symmetry_headers.h"
#include "relation_headers.h"

using namespace std;



// **********************************************************************
// ***               AutomorphismConstraint Class                     ***
// **********************************************************************

AutomorphismConstraint::AutomorphismConstraint
                                   (const set<const FunctionArguments*> &fas,
				    const set<int>& dom,
				    const RelationStore& relstore,
				    bool argument_perms,
				    bool da_left, bool da_right, bool da_both,
				    bool accuracy)
{
  this->domain = dom;
  this->argument_perms = argument_perms;
  this->da_left = da_left;
  this->da_right = da_right;
  this->da_both = da_both;
  this->accuracy = accuracy;

  done_ecg_initialisation = false;

  // Later on, will probably call this from somewhere else.  Whereupon,
  // we'll have to store fas and relstore somewhere in the meantime.  But
  // these are problems for another day.
  initialise_ecg(fas, relstore);
}




// *********************




Constraint::SatLevel AutomorphismConstraint::apply_constraint
                            (Function &f,
			     const set<const FunctionArguments*> &modified_fas,
			     set<const FunctionArguments*> &new_modified_fas,
			     bool complete) const
{
  // If we haven't yet decided on the appropriate tuple ordering, then we
  // quit immediately.  As the program stands at present, this will never
  // occur.  But the code is present to allow for future expansion.
  if (!done_ecg_initialisation)
    return SymmetryBreaking;

  // Loop over each symmetry
  for (ECG::const_iterator fngi = ecg.begin(); fngi != ecg.end(); ++fngi)
  {
    // And over each tuple
    for (nat t = 0; t < fa_ordering.size(); ++t)
    {
      const FunctionArguments* this_fa  = fa_ordering[t];
      const FunctionArguments* other_fa = (fngi->first)[t];
      // Find out the possible values taken by f(tuple t)...
      const FunctionValues* this_fv     = f[this_fa];
      // ... and by f(mapped tuple t)
      const FunctionValues* other_fv = f[other_fa];
      
      // Note that other_fv contains the original (not the mapped) version
      // of the new fn's values.
      // We must now work out what the result of applying the map to the
      // result values is
      int other_min;
      int other_max;
      if ((fngi->second).size() > 0)
      {
	set<int> other_values = other_fv->function_values_to_set();
	DOMAIN_AUTOMORPHISM::const_iterator da_pos =
	  (fngi->second).find(*(other_values.begin()));
	
	if (da_pos == (fngi->second).end())
	  throw SanityError();
	
	other_min = da_pos->second;
	other_max = other_min;
	for (set<int>::const_iterator ovi = other_values.begin();
	     ovi != other_values.end();
	     ++ovi)
	{
	  DOMAIN_AUTOMORPHISM::const_iterator da_pos =
	    (fngi->second).find(*ovi);
	  
	  if (da_pos == (fngi->second).end())
	    throw SanityError();
	  
	  int other_value = da_pos->second;
	  if (other_value < other_min)
	    other_min = other_value;
	  else if (other_value > other_max)
	    other_max = other_value;
	}
      }
      else
      {
	other_min = other_fv->min_value();
	other_max = other_fv->max_value();
      }
      // Find max value of f(tuple t)
      int this_max = this_fv->max_value();
      if (this_max < other_min)
      {
	// Other perm isn't correct representative
	break;
      }

      // Find min value of f(tuple t)
      int this_min = this_fv->min_value();
      if (this_min > other_max)
      {
	// We're not the correct representative
	return NeverSatisfied;
      }
 
      if (this_min == this_max && other_min == other_max)
      {
	// ('&& this_min == other_min' can be deduced.)
	// The two perms are identical on this tuple.
	continue;
      }

      // Finally, if we get here, there's no evidence we're chasing the
      // wrong permutation
      return SymmetryBreaking;
    }
  }
  return AlwaysSatisfied;
}




// *********************




void AutomorphismConstraint::display(ostream& s) const
{
  if (argument_perms || da_left || da_right || da_both)
  {
    s << "f is the canonical function from the equivalence class formed by applying\n";
    char leader[] = "   * ";

    if ((nat) argument_perms +
	(nat) da_left +
	(nat) da_right +
	(nat) da_both == 1)   // Only one can ever be true
    {
      leader[0] = 0;
    }
    
    if (argument_perms)
      s << leader << "permutations to its arguments\n";
    
    if (da_left)
      s << leader << "automorphisms to the left of the function\n";
    
    if (da_right)
      s << leader << "automorphisms to the right of the function\n";
    
    if (da_both)
      s << leader << "automorphisms to both sides of the function simultaneously\n";
    
    if (!accuracy)
    {
      s << "(We prefer speed over accuracy here: some pairs of essentially identical\n";
      s << "functions may both be listed.\n";
    }
  }
  else
  {
    s << "(An empty symmetry-removing automorphism constraint)\n";
  }
  s << "\n";
}




// *********************




void AutomorphismConstraint::initialise_ecg
                               (const set<const FunctionArguments*> &fas,
				const RelationStore& relstore) const
{
  // Set up the normal order of FunctionArguments
  for (FASI fasi = fas.begin(); fasi != fas.end(); ++fasi)
  {
    // If we are just permuting arguments, we can ignore constant FAs, since
    // arg. perms. don't affect them.  This gives a slight speed-up.
    if (argument_perms && !da_left && !da_right && !da_both)
    {
      if (!(*fasi)->is_constant_function_arguments())
	fa_ordering.push_back(*fasi);
    }
    else
    {
      fa_ordering.push_back(*fasi);
    }
  }
  
  DOMAIN_AUTOMORPHISM identity_automorphism;
  
  for (set<int>::const_iterator di = domain.begin(); di != domain.end(); ++di)
    identity_automorphism[*di] = *di;
  
  FN_GENERATOR identity_fn_gen = make_pair(fa_ordering, identity_automorphism);

  if (ecg.size() != 0)
  {
    ecg.clear();
    cerr << "Warning: ecg has a non-zero size.  This could indicate a programming error.\n";
  }

  // If we are to apply any autmorphisms other than the argument permutation,
  // we need to know what the injective unary functions of the relation store
  // are.
  set<DOMAIN_AUTOMORPHISM> iufs;
  if (da_left || da_right || da_both)
    relstore.get_injective_unary_polymorphisms(&iufs);

  // Repeatedly ask all the necessary automorphism symmetries to add new
  // function generators to the set of function generators (ecg) we've
  // already accumulated.
  ECG additional_fn_gens;     // Fn generators we are told to add by a symmetry
  ECG new_additional_fn_gens; // A subset: those which are genuinely novel

  new_additional_fn_gens.insert(identity_fn_gen);

  int iterations = 0;
  do
  {
    iterations++;
    additional_fn_gens.clear();

    if (argument_perms)
      add_fn_gens_argument_perms(new_additional_fn_gens, additional_fn_gens);
    
    if (da_left)
      add_fn_gens_da_left(new_additional_fn_gens, iufs, additional_fn_gens);

    if (da_right)
      add_fn_gens_da_right(new_additional_fn_gens, iufs, additional_fn_gens);

    if (da_both)
      add_fn_gens_da_both(new_additional_fn_gens, iufs, additional_fn_gens);

    // ecg = ecg \union new_additional_fn_gens
    ECG temp_ecg;
    set_union(ecg.begin(),
	      ecg.end(),
	      new_additional_fn_gens.begin(),
	      new_additional_fn_gens.end(),
	      inserter(temp_ecg, temp_ecg.begin()));
    ecg.swap(temp_ecg);
    
    // new_additional_function_gens = additional_function_gens \ ecg
    new_additional_fn_gens.clear();
    set_difference(additional_fn_gens.begin(),
		   additional_fn_gens.end(),
		   ecg.begin(),
		   ecg.end(),
		   inserter(new_additional_fn_gens,
			    new_additional_fn_gens.begin()));
    
  } while ((accuracy && new_additional_fn_gens.size() > 0) ||
	   (!accuracy && iterations < 2));

  // Remove the identity function generator from ecg
  if (ecg.erase(identity_fn_gen) == 0)
  {
    cerr << "Error: Programming error.  The identity function generator for automorphism\n";
    cerr << "       symmetries has vanished.\n";
    throw SanityError();
  }

  // Any function generators which have the identity domain automorphism,
  // have their automorphism replaced by the empty automorphism.  This will
  // be detected in apply_constraint(), and is for speed.
  DOMAIN_AUTOMORPHISM empty_automorphism;
  empty_automorphism.clear();
  ECG temp_ecg;
  
  for (ECG::const_iterator fngi = ecg.begin(); fngi != ecg.end(); ++fngi)
  {
    FN_GENERATOR fn_gen = *fngi;
    if (fn_gen.second == identity_automorphism)
      fn_gen.second = empty_automorphism;
    
    temp_ecg.insert(fn_gen);
  }
  ecg.swap(temp_ecg);

  cout << "(Size of group we're factoring out: " << ecg.size() + 1 << ")\n";
  
  done_ecg_initialisation = true;
}




// ********************




void AutomorphismConstraint::compose_automorphisms
                             (const DOMAIN_AUTOMORPHISM& a1,
			      const DOMAIN_AUTOMORPHISM& a2,
			      DOMAIN_AUTOMORPHISM& result) const
{
  result.clear();
  
  for (set<int>::const_iterator di = domain.begin(); di != domain.end(); ++di)
  {
    int dom_element = *di;
    DOMAIN_AUTOMORPHISM::const_iterator a2_app = a2.find(dom_element);
    
    if (a2_app == a2.end())
      throw SanityError();

    // Find the result of applying a2
    int a2_result = a2_app->second;
    
    DOMAIN_AUTOMORPHISM::const_iterator a1_app = a1.find(a2_result);
    
    if (a1_app == a1.end())
      throw SanityError();

    // Find the result of applying a1
    int a1_result = a1_app->second;
    
    result[dom_element] = a1_result;
  }
}




// ******************




void AutomorphismConstraint::apply_automorphism_to_fa_vec
                             (const DOMAIN_AUTOMORPHISM& da,
			      const FA_VEC& fa_vec,
			      FA_VEC& result) const
{
  result.clear();
  
  for (FA_VEC::const_iterator favi = fa_vec.begin();
       favi != fa_vec.end();
       ++favi)
  {
    vector<int> this_fa_as_vec = (*favi)->function_arguments_to_tuple();
    
    for (nat i = 0; i < this_fa_as_vec.size(); ++i)
    {
      DOMAIN_AUTOMORPHISM::const_iterator da_pos = da.find(this_fa_as_vec[i]);
      
      if (da_pos == da.end())
	throw SanityError();
      
      this_fa_as_vec[i] = da_pos->second;
    }
    result.push_back(
	      FunctionArguments::tuple_to_function_arguments(this_fa_as_vec));
  }
}




// ******************




// Here are the four functions which define what the different symmetries
// actually are.  They are called by the main constraint as and when needed.

void AutomorphismConstraint::add_fn_gens_argument_perms
                             (const ECG fn_gens,
			      ECG& additional_fn_gens) const
{
  for (ECG::const_iterator fngi = fn_gens.begin();
       fngi != fn_gens.end();
       ++fngi)
  {
    nat number_of_fas = (fngi->first).size();

    // There are no permutations of the null function
    if (number_of_fas == 0)
      return;
    
    nat number_of_perms = ((fngi->first)[0])->get_number_of_permutations();

    DOMAIN_AUTOMORPHISM da = fngi->second;
    for (nat perm = 0; perm < number_of_perms; ++perm)  // For each permutation
    {
      FA_VEC fa_ordering;
      for (nat fa = 0; fa < number_of_fas; ++fa)        // For each tuple
	fa_ordering.push_back(((fngi->first)[fa])->get_permutation(perm));
      
      additional_fn_gens.insert(make_pair(fa_ordering, da));
    }
  }
}




// ******************




void AutomorphismConstraint::add_fn_gens_da_left
                             (const ECG& fn_gens,
			      const set<DOMAIN_AUTOMORPHISM>& iufs,
			      ECG& additional_fn_gens) const
{
  for (ECG::const_iterator fngi = fn_gens.begin();
       fngi != fn_gens.end();
       ++fngi)
  {
    for (set<DOMAIN_AUTOMORPHISM>::const_iterator dai = iufs.begin();
	 dai != iufs.end();
	 ++dai)
    {
      FA_VEC new_fa_vec;
      apply_automorphism_to_fa_vec(*dai, fngi->first, new_fa_vec);
      additional_fn_gens.insert(make_pair(new_fa_vec, fngi->second));
    }
  }
}




// ******************




void AutomorphismConstraint::add_fn_gens_da_right
                             (const ECG& fn_gens,
			      const set<DOMAIN_AUTOMORPHISM>& iufs,
			      ECG& additional_fn_gens) const

{
  for (ECG::const_iterator fngi = fn_gens.begin();
       fngi != fn_gens.end();
       ++fngi)
  {
    for (set<DOMAIN_AUTOMORPHISM>::const_iterator dai = iufs.begin();
	 dai != iufs.end();
	 ++dai)
    {
      DOMAIN_AUTOMORPHISM new_da;
      compose_automorphisms(*dai, fngi->second, new_da);
      additional_fn_gens.insert(make_pair(fngi->first, new_da));
    }
  }
}




// ******************




void AutomorphismConstraint::add_fn_gens_da_both
                             (const ECG& fn_gens,
			      const set<DOMAIN_AUTOMORPHISM>& iufs,
			      ECG& additional_fn_gens) const
{
  for (ECG::const_iterator fngi = fn_gens.begin();
       fngi != fn_gens.end();
       ++fngi)
  {
    for (set<DOMAIN_AUTOMORPHISM>::const_iterator dai = iufs.begin();
	 dai != iufs.end();
	 ++dai)
    {
      FA_VEC new_fa_vec;
      apply_automorphism_to_fa_vec(*dai, fngi->first, new_fa_vec);
      
      DOMAIN_AUTOMORPHISM new_da;
      compose_automorphisms(*dai, fngi->second, new_da);
      additional_fn_gens.insert(make_pair(new_fa_vec, new_da));
    }
  }
}







// *****************************************************************
// ***              EssentiallyNAryConstraint class              ***
// *****************************************************************


// Ensures that all functions found use all their arguments

EssentiallyNAryConstraint::EssentiallyNAryConstraint
                           (const set<const FunctionArguments*>& fas,
			    const set<int>& dom, nat a)
{
  arity = a;
  domain = dom;
  
  // Set up the modifications variable.  Slow, but only done once.
  for (FASI fasi = fas.begin(); fasi != fas.end(); ++fasi)
  {
    vector<list<const FunctionArguments*> > this_fa_mods(arity);
    vector <int> tuple = (*fasi)->function_arguments_to_tuple();
    for (nat i = 0; i < arity; ++i)
    {
      int orig_tuple_i = tuple[i];
      for (set<int>::const_iterator di = domain.begin();
	   di != domain.end();
	   ++di)
      {
	if (*di == orig_tuple_i)
	  continue;
	
	tuple[i] = *di;
	const FunctionArguments* new_fa =
	  FunctionArguments::tuple_to_function_arguments(tuple);
	
	this_fa_mods[i].push_back(new_fa);
      }
      tuple[i] = orig_tuple_i;
    }
    modifications[*fasi] = this_fa_mods;
  }
}




// *********************




void EssentiallyNAryConstraint::check_consistency
                             (nat a, const set<int>& dom,
			      const set<const FunctionArguments*>& fas) const
{
  // Do nothing
}




// *********************




Constraint::SatLevel EssentiallyNAryConstraint::apply_constraint
                       (Function &f,
			const set<const FunctionArguments*> &modified_fas,
			set<const FunctionArguments*> &new_modified_fas,
			bool complete) const
{
  for (FASI fasi = modified_fas.begin(); fasi != modified_fas.end(); ++fasi)
  {
    const FunctionValues* original_fv = f[*fasi];

    map<const FunctionArguments*,
      vector<list<const FunctionArguments*> > >::const_iterator
      this_fa_position = modifications.find(*fasi);
    
    if (this_fa_position == modifications.end())
      throw SanityError();
    
    const vector<list<const FunctionArguments*> >& this_fa_mods =
      this_fa_position->second;

    // a for-loop in disguise
    list<nat>::iterator li = undetermined_fn_posns.begin();
    while (li != undetermined_fn_posns.end())
    {
      const FunctionValues* current_fv = original_fv;
      const list<const FunctionArguments*>& this_fa_index_mods =
	this_fa_mods[*li];
      
      list<nat>::iterator next_li = ++li;
      
      // Warning: can't directly dereference li from here on down, because
      // we've just changed its value.  If we want to defererence the former
      // value of li, we'll have to decrease it first.  We have stored
      // away li's successor, so that we can carry on iterating through
      // this "for loop".
      
      for (list<const FunctionArguments*>::const_iterator fali =
	     this_fa_index_mods.begin();
	   fali != this_fa_index_mods.end();
	   ++fali)
      {
	current_fv = current_fv->intersect(f[*fali]);
	if (current_fv->is_empty_set())
	{
	  // Remove *li from the list of undetermined_fn_posns, and go on
	  // to next undetermined_fn_posn
	  essential_fn_posns[depth].splice(essential_fn_posns[depth].begin(),
					   undetermined_fn_posns,
					   --li);
	  break;
	}
      }
      // 'break' takes us to here.
      li = next_li;
    }
  }

  if (undetermined_fn_posns.size() == 0)
    return AlwaysSatisfied;
  
  if (!complete)                     // Can we do better than !complete?
    return SymmetryBreaking;
  
  return NeverSatisfied;
}




// *********************




void EssentiallyNAryConstraint::notify_search_starting(nat max_d)
{
  depth = 0;
  max_depth = max_d;

  // Resize undetermined_fn_posns, and make sure it's empty
  essential_fn_posns.resize(max_depth+1);
  for (nat i = 0; i < max_depth + 1; ++i)
  {
    if (essential_fn_posns[i].size() != 0)
      throw SanityError();
  }
  
  // Initially, we don't know whether any function positions are used
  // essentially
  
  if (undetermined_fn_posns.size() != 0)
    throw SanityError();
  
  for (nat i = 0; i < arity; ++i)
    undetermined_fn_posns.push_back(i);
}




// *********************




void EssentiallyNAryConstraint::notify_search_descent()
{
  depth++;
}



// *********************



void EssentiallyNAryConstraint::notify_search_ascent()
{
  // Put back any indices which we had said were essential
  undetermined_fn_posns.splice(undetermined_fn_posns.begin(),
			       essential_fn_posns[depth]);

  depth--;
}

void EssentiallyNAryConstraint::notify_search_ending()
{
  undetermined_fn_posns.clear();
}    




// *********************




void EssentiallyNAryConstraint::display(ostream& s) const
{
  s << "Every argument of every function is used\n";
}
